# -*- coding: utf-8 -*-
import datetime
import os
import time
import unittest
from io import StringIO
from unittest import mock
import requests
from requests.exceptions import RetryError
from urllib3.exceptions import NewConnectionError

import tos
from tos import DnsCacheService, RateLimiter, exceptions, utils, convert_storage_class_type, ACLType, convert_acl_type, \
    StorageClassType, MetadataDirectiveType, convert_metadata_directive_type, AzRedundancyType, \
    convert_az_redundancy_type, PermissionType, convert_permission_type, GranteeType, convert_grantee_type, \
    convert_canned_type, CannedType, RedirectType, convert_redirect_type, StatusType, convert_status_type, \
    StorageClassInheritDirectiveType, convert_storage_class_inherit_directive_type, VersioningStatusType, \
    convert_versioning_status_type, ProtocolType, convert_protocol_type, CertStatus, convert_cert_status, \
    StaticCredentialsProvider
from tos.auth import CredentialProviderAuth
from tos.checkpoint import CancelHook
from tos.clientv2 import _handler_retry_policy, _is_wrapper_data, _signed_req
from tos.exceptions import TosServerError, CancelNotWithAbortError, CancelWithAbortError
from tos.http import Response, Request
from tos.utils import SizeAdapter


class BaseFuncTestCase(unittest.TestCase):
    def __init__(self, *args, **kwargs):
        super(BaseFuncTestCase, self).__init__(*args, **kwargs)
        self.ak = os.getenv('AK')
        self.sk = os.getenv('SK')
        self.endpoint = os.getenv('Endpoint')
        self.region = os.getenv('Region')

    def test_cache(self):
        expire = 2
        cache = DnsCacheService()
        port = 8080
        list_ip = [['1', 1], ['2', 2]]

        cache.add('baidu.com', port, list_ip, expire=int(time.time()) + expire)
        cache.add('baidu.com2', port, list_ip, expire=int(time.time()) + expire)
        entry = cache.get_ip_list('baidu.com', port)
        self.assertIsNotNone(entry)
        time.sleep(1)
        self.assertIsNotNone(cache.get_ip_list('baidu.com', port))
        time.sleep(2)
        self.assertIsNone(cache.get_ip_list('baidu.com', port))

    def test_retry(self):
        with open('out.txt', 'wb') as f:
            f.write(b'123')

        for t in [
            args('', 'GET', '', requests.Timeout(), None, True),
            args('', 'HEAD', '', requests.Timeout(), None, True),
            args('', 'HEAD', '', requests.ConnectionError(), None, True),
            args('', 'HEAD', '', RetryError(), None, True),
            args('', 'PUT', '', requests.Timeout(), None, True),
            args('', 'POST', '', RetryError(), None, True),
            args('', 'POST', '', RetryError(), None, True),
            args('', 'GET', '', NewConnectionError(None, None), None, True),
            args('', 'PUT', 'put_object_acl', None, TestException(429), True),
            args('', 'PUT', 'put_object_acl', None, TestException(501), True),
            args('', 'PUT', 'put_object_acl', None, TestException(301), False),
            args('', 'PUT', 'create_bucket', None, TestException(429), True),
            args('', 'PUT', 'create_bucket', None, TestException(301), False),
            args('', 'DELETE', 'delete_bucket', None, TestException(429), True),
            args('', 'DELETE', 'delete_bucket', None, TestException(301), False),
            args('123', 'PUT', 'create_bucket', None, TestException(429), True),
            args('123', 'PUT', 'upload_part', None, TestException(429), True),

            # 测试 字符串 reset
            args(wrapper_data(
                tos.utils.init_content(SizeAdapter(open('out.txt', 'rb'), 2, init_offset=1, can_reset=True)),
                True, True, True), 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'23'),

            args(wrapper_data(tos.utils.init_content('123'), True, False, False),
                 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'123'),

            args(wrapper_data(tos.utils.init_content('123'), False, True, False),
                 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'123'),

            args(wrapper_data(tos.utils.init_content('123'), False, False, True),
                 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'123'),

            args(wrapper_data(tos.utils.init_content('123'), True, True, True),
                 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'123'),

            args(wrapper_data(tos.utils.init_content('123'), False, True, True),
                 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'123'),

            # 测试 StringIO reset
            args(wrapper_data(tos.utils.init_content(StringIO('123'), init_offset=0), False, False, False),
                 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'123'),

            args(wrapper_data(tos.utils.init_content(StringIO('123'), init_offset=0), True, True, True),
                 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'123'),

            # 测试 StringIO reset 偏移
            args(wrapper_data(tos.utils.init_content(StringIO('123'), init_offset=1), False, False, False),
                 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'23'),

            args(wrapper_data(tos.utils.init_content(StringIO('123'), init_offset=1), True, True, True),
                 'PUT', 'create_bucket', None, TestException(429), True, expect_data=b'23'),
        ]:
            if _is_wrapper_data(data=t.body):
                self.assertTrue(t.body.read(1) != t.body.read(1))
                if t.body.crc_callback:
                    self.assertTrue(t.body.crc != 0)

            got = _handler_retry_policy(t.body, t.method, t.fun_name, t.client_exp, t.server_exp)
            self.assertEqual(t.want, got)

            if _is_wrapper_data(data=t.body):
                if t.body.crc_callback:
                    self.assertTrue(t.body.crc == 0)
            if t.expect_data and _is_wrapper_data(data=t.body):
                self.assertEqual(t.expect_data, t.body.read())
                if t.body.crc_callback:
                    self.assertTrue(t.body.crc != 0)

        os.remove('out.txt')

    def test_cancel_hook(self):
        cancle = CancelHook()
        cancle.cancel(False)
        with self.assertRaises(CancelNotWithAbortError):
            cancle.is_cancel()
        cancle.cancel(True)
        with self.assertRaises(CancelWithAbortError):
            cancle.is_cancel()

    def test_stop_check(self):
        io = StringIO("123")
        req = TestResponse(4, io)
        with self.assertRaises(tos.exceptions.TosClientError):
            req.read()

        io = StringIO("123")
        req = TestResponse(4, io)
        with self.assertRaises(tos.exceptions.TosClientError):
            for content in req:
                print(content)

        io = StringIO("123")
        req = TestResponse(3, io)
        for content in req:
            print(content)

    def test_with_endpint(self):
        client = tos.TosClientV2(self.ak, self.sk, self.endpoint, self.region)
        self.assertEqual(client.endpoint, "https://" + self.endpoint)
        self.assertEqual(client.scheme, "https://")
        self.assertEqual(client.host, self.endpoint)

        client_2 = tos.TosClientV2(self.ak, self.sk, 'http://' + self.endpoint, self.region)
        self.assertEqual(client_2.endpoint, "http://" + self.endpoint)
        self.assertEqual(client_2.host, self.endpoint)
        self.assertEqual(client_2.scheme, "http://")

        client = tos.TosClientV2(self.ak, self.sk, "", "cn-beijing")
        self.assertEqual(client.endpoint, 'https://' + 'tos-cn-beijing.volces.com')
        self.assertEqual(client.host, 'tos-cn-beijing.volces.com')
        self.assertEqual(client.scheme, 'https://')

        client = tos.TosClientV2(self.ak, self.sk, "tos-cn-beijing.ivolces.com", "cn-beijing")
        self.assertEqual(client.endpoint, 'https://' + 'tos-cn-beijing.ivolces.com')
        self.assertEqual(client.host, 'tos-cn-beijing.ivolces.com')
        self.assertEqual(client.scheme, 'https://')

    def test_with_invalid_endpint(self):
        with self.assertRaises(tos.exceptions.TosClientError):
            tos.TosClientV2(self.ak, self.sk, 'tos-s3-cn-beijing.volces.com', self.region)

    def test_convert_enum_type(self):
        for t in ACLType:
            assert t == convert_acl_type(t.value)
        assert ACLType.ACL_Unknown == convert_acl_type('test')

        for t in StorageClassType:
            assert t == convert_storage_class_type(t.value)
        assert StorageClassType.Storage_Unknown == convert_storage_class_type('test')

        for t in MetadataDirectiveType:
            assert t == convert_metadata_directive_type(t.value)
        assert MetadataDirectiveType.Metadata_Directive_Unknown == convert_metadata_directive_type('test')

        for t in AzRedundancyType:
            assert t == convert_az_redundancy_type(t.value)
        assert AzRedundancyType.Az_Redundancy_Unknown == convert_az_redundancy_type('test')

        for t in PermissionType:
            assert t == convert_permission_type(t.value)
        assert PermissionType.Permission_Unknown == convert_permission_type('test')

        for t in GranteeType:
            assert t == convert_grantee_type(t.value)
        assert GranteeType.Grantee_Unknown == convert_grantee_type('test')

        for t in CannedType:
            assert t == convert_canned_type(t.value)
        assert CannedType.Canned_Unknown == convert_canned_type('test')

        for t in RedirectType:
            assert t == convert_redirect_type(t.value)
        assert RedirectType.Unknown == convert_redirect_type('test')

        for t in StatusType:
            assert t == convert_status_type(t.value)
        assert StatusType.Status_Unknown == convert_status_type('test')

        for t in StorageClassInheritDirectiveType:
            assert t == convert_storage_class_inherit_directive_type(t.value)
        assert StorageClassInheritDirectiveType.Storage_Class_Unknown == convert_storage_class_inherit_directive_type(
            'test')

        for t in VersioningStatusType:
            assert t == convert_versioning_status_type(t.value)
        assert VersioningStatusType.Versioning_Unknown == convert_versioning_status_type('test')

        for t in ProtocolType:
            assert t == convert_protocol_type(t.value)
        assert ProtocolType.Protocol_Unknown == convert_protocol_type('test')

        for t in CertStatus:
            assert t == convert_cert_status(t.value)
        assert CertStatus.Cert_Unknown == convert_cert_status('test')

    def test_sign_req(self):
        auth = CredentialProviderAuth(StaticCredentialsProvider('ak', 'sk', 'sts'), 'region')
        host = 'zzz.com'
        headers = {'content-type': 'application/json', 'Host': host}
        params = {'versionId': 'test'}

        req = Request('GET', 'http://zzz.com/key', 'key', 'zzz.com',
                      params=params,
                      headers=headers)

        datetime_mock = mock.Mock(wraps=datetime.datetime)
        datetime_mock.utcnow.return_value = datetime.datetime(2021, 1, 1)
        with mock.patch('datetime.datetime', new=datetime_mock):
            req = _signed_req(auth, req, host)
            self.assertEqual(req.headers['Authorization'],
                             'TOS4-HMAC-SHA256 Credential=ak/20210101/region/tos/request, '
                             'SignedHeaders=content-type;host;x-tos-date;x-tos-security-token, Signature=ca8eb8987663e61e740bc5be9078d6d4f716990573a56c86a6602d42faf3af7e')
            self.assertEqual(req.headers['Host'], host)
            self.assertEqual(req.headers['x-tos-date'], '20210101T000000Z')
            self.assertEqual(req.headers['x-tos-security-token'], 'sts')


class args(object):
    def __init__(self, body, method, fun_name, client_exp, server_exp, want, expect_data=None):
        self.body = body
        self.method = method
        self.fun_name = fun_name
        self.client_exp = client_exp
        self.server_exp = server_exp
        self.want = want
        self.expect_data = expect_data


class TestException(TosServerError):
    def __init__(self, code):
        self.status_code = code

    def __str__(self):
        return str(self.status_code)


def wrapper_data(date, wrapper_progress, wrapper_limiter, wrapper_crc):
    def progress(consumed_bytes, total_bytes, rw_once_bytes,
                 type):
        print("consumed_bytes:{0},total_bytes{1}, rw_once_bytes:{2}, type:{3}".format(consumed_bytes, total_bytes,
                                                                                      rw_once_bytes, type))

    limiter = RateLimiter(1200, 10000)

    if wrapper_progress:
        date = tos.utils.add_progress_listener_func(date, progress_callback=progress)

    if wrapper_limiter:
        date = tos.utils.add_rate_limiter_func(date, rate_limiter=limiter)

    if wrapper_crc:
        date = tos.utils.add_crc_func(date, init_crc=0)

    return date


class TestResponse(Response):
    def __init__(self, length, body):
        self.content_length = length
        self.resp = body
        self._all_read = False
        self.offset = 0

    def read(self, amt=None):
        if self._all_read:
            return b''

        if amt is None:
            content = self.resp.read()

            self._all_read = True
            if self.content_length and len(content) != self.content_length:
                raise tos.exceptions.TosClientError('IO Content not equal content-length')
            return content
        else:
            try:
                read = next(self.resp)
                self.offset += len(read)
                return read
            except StopIteration:
                if self.content_length and self.offset != self.content_length:
                    raise exceptions.TosClientError('IO Content not equal content-length')
                self._all_read = True
                return b''


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-

import datetime
import unittest
from unittest import mock

import tos
from tests.common import TosTestCase
from tos import HttpMethodType
from tos.models2 import PostSignatureCondition, ContentLengthRange


class TestAuth(TosTestCase):
    maxDiff = None

    def test_generate_presigned_url(self):
        datetime_mock = mock.Mock(wraps=datetime.datetime)
        datetime_mock.utcnow.return_value = datetime.datetime(2021, 1, 1)
        with mock.patch('datetime.datetime', new=datetime_mock):
            tos_cli = tos.TosClient(tos.Auth('ak', 'sk', 'beijing'), 'tos-cn-beijing.volces.com')
            url = tos_cli.generate_presigned_url(Method='GET', Bucket='bkt', Key='key', ExpiresIn=86400)
            self.assertEqual(url,
                             'https://bkt.tos-cn-beijing.volces.com/key?X-Tos-Algorithm=TOS4-HMAC-SHA256&X-Tos-Creden'
                             'tial=ak%2F20210101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20210101T000000Z&X-Tos-Expires=8'
                             '6400&X-Tos-SignedHeaders=host&X-Tos-Signature=b87788cb98d1a5a91a046d20eb212ffc22cf7cd'
                             '4c1d4e9bd2d15a989afb97d2f')

            tos_cli = tos.TosClient(tos.Auth('ak', 'sk', 'beijing', sts='sts'), 'tos-cn-beijing.volces.com')
            url = tos_cli.generate_presigned_url(Method='PUT', Bucket='bkt', Key='key')
            self.assertEqual(url,
                             'https://bkt.tos-cn-beijing.volces.com/key?X-Tos-Algorithm=TOS4-HMAC-SHA256&X-Tos-Credentia'
                             'l=ak%2F20210101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20210101T000000Z&X-Tos-Expires=36'
                             '00&X-Tos-Security-Token=sts&X-Tos-SignedHeaders=host&X-Tos-Signature=3041fb481e31ec25f'
                             'e7a1be44cafe25caf1cd97228710ee440b2ca1bd49bc563')

            url = tos_cli.generate_presigned_url(Method='PUT', Bucket='bkt', Key='key', Params={'acl': ''})
            self.assertEqual(url,
                             'https://bkt.tos-cn-beijing.volces.com/key?acl&X-Tos-Algorithm=TOS4-HMAC-SHA256&X-Tos-C'
                             'redential=ak%2F20210101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20210101T000000Z&X-Tos-Exp'
                             'ires=3600&X-Tos-Security-Token=sts&X-Tos-SignedHeaders=host&X-Tos-Signature=51c89070206'
                             'dd438fd8be1ed201a77335af80e33cad5fd408841590b99aa93d3')

    def test_client2_presigned_url(self):
        datetime_mock = mock.Mock(wraps=datetime.datetime)
        datetime_mock.utcnow.return_value = datetime.datetime(2021, 1, 1)
        with mock.patch('datetime.datetime', new=datetime_mock):
            tos_cli = tos.TosClientV2(ak='ak', sk='sk', endpoint='tos-cn-beijing.volces.com', region='beijing')
            url = tos_cli.pre_signed_url(http_method=HttpMethodType.Http_Method_Get, bucket='bkt', key='key',
                                         expires=704800)
            print(url.signed_url)
            self.assertEqual(url.signed_url,
                             'https://bkt.tos-cn-beijing.volces.com/key?X-Tos-Algorithm=TOS4-HMAC-SHA256&X-Tos-Credential=ak%2F20210101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20210101T000000Z&X-Tos-Expires=704800&X-Tos-SignedHeaders=host&X-Tos-Signature=087f3eb174b6accb37178630ba5890d7d8bc8a33495e25959dc359dc4b0f1170')
            'https://bkt.tos-cn-beijing.volces.com/key?X-Tos-Expires=86400&X-Tos-SignedHeaders=host&X-Tos-Credential=ak%2F20210101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20210101T000000Z&X-Tos-Signature=b87788cb98d1a5a91a046d20eb212ffc22cf7cd4c1d4e9bd2d15a989afb97d2f&X-Tos-Algorithm=TOS4-HMAC-SHA256'

            tos_cli = tos.TosClientV2(ak='ak', sk='sk', endpoint='tos-cn-beijing.volces.com1', region='beijing')
            url = tos_cli.pre_signed_url(http_method=HttpMethodType.Http_Method_Get, bucket='bkt', key='key',
                                         expires=86400, alternative_endpoint='tos-cn-beijing.volces.com')
            self.assertEqual(url.signed_url,
                             'https://bkt.tos-cn-beijing.volces.com/key?X-Tos-Algorithm=TOS4-HMAC-SHA256&X-Tos-Creden'
                             'tial=ak%2F20210101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20210101T000000Z&X-Tos-Expires=8'
                             '6400&X-Tos-SignedHeaders=host&X-Tos-Signature=b87788cb98d1a5a91a046d20eb212ffc22cf7cd'
                             '4c1d4e9bd2d15a989afb97d2f')

            tos_cli = tos.TosClientV2(ak='ak', sk='sk', endpoint='tos-cn-beijing.volces.com', region='beijing',
                                      security_token='sts')
            url = tos_cli.pre_signed_url(http_method=HttpMethodType.Http_Method_Put, bucket='bkt', key='key',
                                         query={'acl': ''})

            self.assertEqual(url.signed_url,
                             'https://bkt.tos-cn-beijing.volces.com/key?acl&X-Tos-Algorithm=TOS4-HMAC-SHA256&X-Tos-C'
                             'redential=ak%2F20210101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20210101T000000Z&X-Tos-Exp'
                             'ires=3600&X-Tos-Security-Token=sts&X-Tos-SignedHeaders=host&X-Tos-Signature=51c89070206'
                             'dd438fd8be1ed201a77335af80e33cad5fd408841590b99aa93d3')

            url = tos_cli.pre_signed_url(http_method=HttpMethodType.Http_Method_Put, bucket='bkt', key='key')
            self.assertEqual(url.signed_url,
                             'https://bkt.tos-cn-beijing.volces.com/key?X-Tos-Algorithm=TOS4-HMAC-SHA256&X-Tos-Credentia'
                             'l=ak%2F20210101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20210101T000000Z&X-Tos-Expires=36'
                             '00&X-Tos-Security-Token=sts&X-Tos-SignedHeaders=host&X-Tos-Signature=3041fb481e31ec25f'
                             'e7a1be44cafe25caf1cd97228710ee440b2ca1bd49bc563')
            url = tos_cli.pre_signed_url(http_method=HttpMethodType.Http_Method_Put, bucket='bkt', key='key',
                                         expires=1234, header={'contentLength': 1000})
            self.assertTrue('1234' in url.signed_url)
            self.assertTrue('contentlength' not in url.signed_url)

            self.anonymousCli = tos.TosClientV2(ak='', sk='', endpoint='tos-cn-beijing.volces.com', region='beijing')
            url = self.anonymousCli.pre_signed_url(http_method=HttpMethodType.Http_Method_Put, bucket='bkt', key='key',
                                                   header={'contentLength': 1000}, query={'t1': 't1'})
            self.assertEqual(url.signed_url, 'https://bkt.tos-cn-beijing.volces.com/key?t1=t1')

            url = self.anonymousCli.pre_signed_url(http_method=HttpMethodType.Http_Method_Put, bucket='bkt', key='key')
            self.assertEqual(url.signed_url, 'https://bkt.tos-cn-beijing.volces.com/key?')

    def test_pre_signed_post_signature(self):
        datetime_mock = mock.Mock(wraps=datetime.datetime)
        datetime_mock.utcnow.return_value = datetime.datetime(2022, 1, 1)
        with mock.patch('datetime.datetime', new=datetime_mock):
            tos_cli = tos.TosClientV2(ak='ak', sk='sk', endpoint='tos-cn-beijing.volces.com', region='beijing')
            conditions = [PostSignatureCondition(key='acl', value='public-read'),
                          PostSignatureCondition(key='Content-Type', value='image/', operator='starts-with')]
            # conditions.append(PostSignatureCondition(key="key", value="example", operator="starts-with"))
            content_length_range = ContentLengthRange(range_start=1023, range_end=10000)
            out = tos_cli.pre_signed_post_signature(bucket='testBucket', key='test_object', conditions=conditions,
                                                    expires=704800, content_length_range=content_length_range)

    def test_pre_signed_policy_url(self):
        datetime_mock = mock.Mock(wraps=datetime.datetime)
        datetime_mock.utcnow.return_value = datetime.datetime(2022, 1, 1)
        with mock.patch('datetime.datetime', new=datetime_mock):
            tos_cli = tos.TosClientV2(ak='ak', sk='sk', endpoint='tos-cn-beijing.volces.com', region='beijing')
            conditions = [PostSignatureCondition(key='bucket', value='examplebucket'),
                          PostSignatureCondition(key='key', value='abc/', operator='starts-with'),
                          PostSignatureCondition(key='key', value='aaa/abc/', operator='starts-with'),
                          PostSignatureCondition(key='key', value='exampleobject', operator='eq'),
                          PostSignatureCondition(key='key', value='exampleobject1', operator='eq')]
            out = tos_cli.pre_signed_policy_url(bucket='test', conditions=conditions, expires=704800)
            list_1 = out.get_signed_url_for_list({'k1': 'v1', 'k2': 'v2'})
            self.assertEqual(list_1,
                             'https://test.tos-cn-beijing.volces.com?X-Tos-Algorithm=TOS4-HMAC-SHA256&X-Tos-Credential=ak%2F20220101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20220101T000000Z&X-Tos-Expires=704800&X-Tos-Policy=eyJjb25kaXRpb25zIjogW3siYnVja2V0IjogImV4YW1wbGVidWNrZXQifSwgWyJzdGFydHMtd2l0aCIsICIka2V5IiwgImFiYy8iXSwgWyJzdGFydHMtd2l0aCIsICIka2V5IiwgImFhYS9hYmMvIl0sIFsiZXEiLCAiJGtleSIsICJleGFtcGxlb2JqZWN0Il0sIFsiZXEiLCAiJGtleSIsICJleGFtcGxlb2JqZWN0MSJdLCB7ImJ1Y2tldCI6ICJ0ZXN0In1dfQ%3D%3D&X-Tos-Signature=869b9e678a6a69ce4a1cd02c7c0e333cb8e7a312a1078563dd2619d3daa9f5ee&k1=v1&k2=v2')
            get = out.get_signed_url_for_get_or_head(key='exampleobject', additional_query={'k1': 'v1', 'k2': 'v2'})
            self.assertEqual(get,
                             'https://test.tos-cn-beijing.volces.com/exampleobject?X-Tos-Algorithm=TOS4-HMAC-SHA256&X-Tos-Credential=ak%2F20220101%2Fbeijing%2Ftos%2Frequest&X-Tos-Date=20220101T000000Z&X-Tos-Expires=704800&X-Tos-Policy=eyJjb25kaXRpb25zIjogW3siYnVja2V0IjogImV4YW1wbGVidWNrZXQifSwgWyJzdGFydHMtd2l0aCIsICIka2V5IiwgImFiYy8iXSwgWyJzdGFydHMtd2l0aCIsICIka2V5IiwgImFhYS9hYmMvIl0sIFsiZXEiLCAiJGtleSIsICJleGFtcGxlb2JqZWN0Il0sIFsiZXEiLCAiJGtleSIsICJleGFtcGxlb2JqZWN0MSJdLCB7ImJ1Y2tldCI6ICJ0ZXN0In1dfQ%3D%3D&X-Tos-Signature=869b9e678a6a69ce4a1cd02c7c0e333cb8e7a312a1078563dd2619d3daa9f5ee&k1=v1&k2=v2')

            anonymous_cli = tos.TosClientV2(ak='', sk='', endpoint='tos-cn-beijing.volces.com', region='beijing')
            out = anonymous_cli.pre_signed_policy_url(bucket='test', conditions=conditions)
            list_1 = out.get_signed_url_for_list({'k1': 'v1', 'k2': 'v2'})
            self.assertEqual(list_1,
                             'https://test.tos-cn-beijing.volces.com?X-Tos-Policy=eyJjb25kaXRpb25zIjogW3siYnVja2V0IjogImV4YW1wbGVidWNrZXQifSwgWyJzdGFydHMtd2l0aCIsICIka2V5IiwgImFiYy8iXSwgWyJzdGFydHMtd2l0aCIsICIka2V5IiwgImFhYS9hYmMvIl0sIFsiZXEiLCAiJGtleSIsICJleGFtcGxlb2JqZWN0Il0sIFsiZXEiLCAiJGtleSIsICJleGFtcGxlb2JqZWN0MSJdLCB7ImJ1Y2tldCI6ICJ0ZXN0In0sIHsiYnVja2V0IjogInRlc3QifV19&k1=v1&k2=v2')
            get = out.get_signed_url_for_get_or_head(key='exampleobject', additional_query={'k1': 'v1', 'k2': 'v2'})
            self.assertEqual(get,
                             'https://test.tos-cn-beijing.volces.com/exampleobject?X-Tos-Policy=eyJjb25kaXRpb25zIjogW3siYnVja2V0IjogImV4YW1wbGVidWNrZXQifSwgWyJzdGFydHMtd2l0aCIsICIka2V5IiwgImFiYy8iXSwgWyJzdGFydHMtd2l0aCIsICIka2V5IiwgImFhYS9hYmMvIl0sIFsiZXEiLCAiJGtleSIsICJleGFtcGxlb2JqZWN0Il0sIFsiZXEiLCAiJGtleSIsICJleGFtcGxlb2JqZWN0MSJdLCB7ImJ1Y2tldCI6ICJ0ZXN0In0sIHsiYnVja2V0IjogInRlc3QifV19&k1=v1&k2=v2')


if __name__ == '__main__':
    unittest.main()



import threading

_thread_ctx = threading.local()


def consume_body():
    current = 0
    if hasattr(_thread_ctx, 'body'):
        current = int(_thread_ctx.body)
        del _thread_ctx.body
    return current


def produce_body(val):
    current = int(val)
    if hasattr(_thread_ctx, 'body'):
        current += int(_thread_ctx.body)
    _thread_ctx.body = current



# -*- coding: utf-8 -*-
import base64
import datetime
import unittest

from tests.common import TosTestBase, random_bytes
from tos.enum import ACLType, StorageClassType
from tos.exceptions import TosServerError
from tos.models2 import UploadedPart


class TestMultipart(TosTestBase):
    def test_multipart(self):
        bucket_name = self.bucket_name + "-test-multipart"
        self.bucket_delete.append(bucket_name)
        self.client.create_bucket(bucket_name)
        key = self.random_key('.js')
        mult_out = self.client.create_multipart_upload(bucket_name, key)
        parts = []
        content = random_bytes(5 * 1024 * 1024)
        for i in range(1, 3):
            upload_part_output = self.client.upload_part(bucket=bucket_name, key=key, upload_id=mult_out.upload_id,
                                                         part_number=i, content=content)
            parts.append(UploadedPart(i, upload_part_output.etag))

        list_parts = self.client.list_parts(bucket_name, key, mult_out.upload_id, max_parts=1)
        self.assertEqual(list_parts.bucket, bucket_name)
        self.assertEqual(list_parts.key, key)
        self.assertEqual(list_parts.upload_id, mult_out.upload_id)
        self.assertEqual(list_parts.max_parts, 1)
        self.assertEqual(list_parts.is_truncated, True)
        self.assertIsNotNone(list_parts.next_part_number_marker)
        self.assertEqual(list_parts.storage_class, StorageClassType.Storage_Class_Standard)
        self.assertTrue(list_parts.owner is not None)
        self.assertTrue(len(list_parts.parts), 1)

        part = list_parts.parts[0]
        self.assertTrue(part.part_number is not None)
        self.assertTrue(len(part.etag) > 0)
        self.assertTrue(part.size is not None)
        self.assertTrue(part.last_modified is not None)

        list_parts_2 = self.client.list_parts(bucket_name, key, mult_out.upload_id,
                                              part_number_marker=list_parts.next_part_number_marker)
        self.assertEqual(len(list_parts_2.parts), 1)
        self.assertEqual(list_parts_2.is_truncated, False)

        complete_out = self.client.complete_multipart_upload(bucket_name, key, mult_out.upload_id, parts=parts)
        self.assertEqual(complete_out.bucket, bucket_name)
        self.assertTrue(len(complete_out.key) > 0)
        self.assertTrue(len(complete_out.etag) > 0)
        self.assertTrue(complete_out.hash_crc64_ecma > 0)

        get_out = self.client.get_object(bucket_name, key)
        self.assertEqual(get_out.read(), content + content)

    def test_multipart_with_options(self):
        bucket_name = self.bucket_name + "-test-multipart-with-option"
        self.bucket_delete.append(bucket_name)
        self.client.create_bucket(bucket_name)
        key = self.random_key('.js')
        meta = {'name': 'sunyushan'}
        mult_out = self.client.create_multipart_upload(bucket_name, key, cache_control="Cache-Control",
                                                       content_disposition="test", content_encoding="utf-8",
                                                       expires=datetime.datetime(2023, 1, 1),
                                                       content_language="english", content_type="text",
                                                       acl=ACLType.ACL_Bucket_Owner_Full_Control,
                                                       meta=meta,
                                                       storage_class=StorageClassType.Storage_Class_Ia
                                                       )
        parts = []
        content = random_bytes(5 * 1024 * 1024)
        for i in range(1, 3):
            upload_part_output = self.client.upload_part(bucket=bucket_name, key=key, upload_id=mult_out.upload_id,
                                                         part_number=i, content=content)
            parts.append(UploadedPart(i, upload_part_output.etag))

        self.client.list_parts(bucket_name, key, mult_out.upload_id)

        self.client.complete_multipart_upload(bucket_name, key, mult_out.upload_id, parts=parts)

        get_out = self.client.get_object(bucket_name, key)
        self.assertTrue(len(get_out.cache_control) > 0)
        self.assertEqual(get_out.content_disposition, "test")
        self.assertEqual(get_out.content_encoding, "utf-8")
        self.assertTrue(get_out.expires is not None)
        self.assertEqual(get_out.content_language, 'english')
        self.assertTrue(get_out.content_type, 'text')
        self.assertEqual(get_out.storage_class, StorageClassType.Storage_Class_Ia)

    def test_multipart_copy(self):
        src_bucket_name = self.bucket_name + "-test-multipart-copy"
        self.bucket_delete.append(src_bucket_name)
        self.client.create_bucket(src_bucket_name)
        key = self.random_key('.js')
        content = random_bytes(1024 * 1024 * 5)

        save_bucket_name = self.bucket_name + "test-multipart" + "v2"
        self.bucket_delete.append(save_bucket_name)
        self.client.create_bucket(save_bucket_name)

        self.client.put_object(src_bucket_name, key, content=content)

        out = self.client.create_multipart_upload(save_bucket_name, key)
        parts = []
        part_copy_1 = self.client.upload_part_copy(save_bucket_name, key, out.upload_id, part_number=1,
                                                   src_bucket=src_bucket_name, src_key=key)
        self.assertIsNotNone(part_copy_1.etag)
        self.assertIsNotNone(part_copy_1.last_modified)
        parts.append(UploadedPart(1, part_copy_1.etag))
        part_copy_2 = self.client.upload_part_copy(save_bucket_name, key, out.upload_id, part_number=2,
                                                   src_bucket=src_bucket_name, src_key=key)
        parts.append(UploadedPart(2, part_copy_2.etag))

        self.client.complete_multipart_upload(save_bucket_name, key, out.upload_id, parts)

        get_out = self.client.get_object(save_bucket_name, key)
        self.assertEqual(get_out.read(), content + content)

    def test_multipart_abort(self):
        bucket_name = self.bucket_name + "-test-multipart-abort"
        self.bucket_delete.append(bucket_name)
        self.client.create_bucket(bucket_name)
        key = self.random_key('.js')
        mult_out = self.client.create_multipart_upload(bucket_name, key)
        parts = []
        content = random_bytes(5 * 1024 * 1024)
        for i in range(1, 3):
            upload_part_output = self.client.upload_part(bucket=bucket_name, key=key, upload_id=mult_out.upload_id,
                                                         part_number=i, content=content)
            parts.append(UploadedPart(i, upload_part_output.etag))

        self.client.abort_multipart_upload(bucket_name, key, mult_out.upload_id)
        with self.assertRaises(TosServerError):
            self.client.list_parts(bucket_name, key, mult_out.upload_id)

    def test_upload_part_from_file(self):
        bucket_name = self.bucket_name + "-test-multipart-file"
        self.bucket_delete.append(bucket_name)
        self.client.create_bucket(bucket_name)
        key = self.random_key('.js')
        mult_out = self.client.create_multipart_upload(bucket_name, key)
        parts = []
        file_name = self.random_filename()
        content = random_bytes(1024 * 1024 * 10)
        with open(file_name, 'wb') as fw:
            fw.write(content)
        for i in range(1, 3):
            upload_part_output = self.client.upload_part_from_file(bucket=bucket_name, key=key,
                                                                   upload_id=mult_out.upload_id,
                                                                   part_number=i, file_path=file_name)
            parts.append(UploadedPart(i, upload_part_output.etag))

        list_parts = self.client.list_parts(bucket_name, key, mult_out.upload_id)
        self.assertEqual(list_parts.bucket, bucket_name)
        self.assertEqual(list_parts.key, key)
        self.assertEqual(list_parts.upload_id, mult_out.upload_id)
        self.assertEqual(list_parts.storage_class, StorageClassType.Storage_Class_Standard)
        self.assertTrue(list_parts.owner is not None)

        part = list_parts.parts[0]
        self.assertTrue(part.part_number is not None)
        self.assertTrue(len(part.etag) > 0)
        self.assertTrue(part.size is not None)
        self.assertTrue(part.last_modified is not None)

        complete_out = self.client.complete_multipart_upload(bucket_name, key, mult_out.upload_id, parts=parts)
        self.assertEqual(complete_out.bucket, bucket_name)
        self.assertTrue(len(complete_out.key) > 0)
        self.assertTrue(len(complete_out.etag) > 0)
        self.assertTrue(complete_out.hash_crc64_ecma > 0)

        self.assertObjectContent(bucket=bucket_name, key=key, content=content + content)

    def test_upload_part_from_file_with_offset(self):
        bucket_name = self.bucket_name + "-test-multipart-file"
        self.bucket_delete.append(bucket_name)
        self.client.create_bucket(bucket_name)
        key = self.random_key('.js')
        mult_out = self.client.create_multipart_upload(bucket_name, key)
        parts = []
        file_name = self.random_filename()
        content = random_bytes(1024 * 13 * 1024)
        with open(file_name, 'wb') as fw:
            fw.write(content)
        for i in range(1, 4):
            upload_part_output = self.client.upload_part_from_file(bucket=bucket_name, key=key,
                                                                   upload_id=mult_out.upload_id,
                                                                   part_number=i, file_path=file_name,
                                                                   offset=5 * (i - 1) * 1024 * 1024,
                                                                   part_size=5 * 1024 * 1024)
            parts.append(UploadedPart(i, upload_part_output.etag))

        self.client.list_parts(bucket_name, key, mult_out.upload_id)
        self.client.complete_multipart_upload(bucket_name, key, mult_out.upload_id, parts=parts)
        self.assertObjectContent(bucket_name, key, content)

    def test_list_multipart_upload(self):
        bucket_name = self.bucket_name + '-test-list'
        self.bucket_delete.append(bucket_name)
        self.client.create_bucket(bucket_name)
        self.client.create_multipart_upload(bucket_name, 'fun1')
        list_all = self.client.list_multipart_uploads(bucket_name)
        self.assertIsNotNone(list_all.encoding_type)
        self.assertTrue(len(list_all.uploads))
        self.assertEqual(list_all.max_uploads, 1000)
        self.assertEqual(list_all.uploads[0].storage_class, StorageClassType.Storage_Class_Standard)
        self.assertIsNotNone(list_all.uploads[0].upload_id)
        self.assertIsNotNone(list_all.uploads[0].owner.id)
        self.assertEqual(list_all.uploads[0].key, 'fun1')
        self.assertIsNotNone(list_all.uploads[0].initiated)

        self.client.create_multipart_upload(bucket_name, 'fun2/')
        self.client.create_multipart_upload(bucket_name, 'fun2/test')
        out_prefix = self.client.list_multipart_uploads(bucket_name, prefix='fun2')
        self.assertTrue(len(out_prefix.uploads), 2)
        self.assertEqual(out_prefix.prefix, 'fun2')

        out_max_keys = self.client.list_multipart_uploads(bucket_name, 'fun2', max_uploads=1)
        self.assertEqual(out_max_keys.max_uploads, 1)
        self.assertTrue(out_max_keys.is_truncated)
        self.assertIsNotNone(out_max_keys.next_key_marker)

        out_max_keys_2 = self.client.list_multipart_uploads(bucket_name, 'fun2/',
                                                            key_marker=out_max_keys.next_key_marker)
        self.assertEqual(out_max_keys_2.key_marker, out_max_keys.next_key_marker)
        self.assertEqual(out_max_keys_2.is_truncated, False)
        self.assertEqual(len(out_max_keys_2.uploads), 1)

        self.client.create_multipart_upload(bucket_name, 'fun3/')
        out_delimiter = self.client.list_multipart_uploads(bucket_name, delimiter='/')
        self.assertEqual(out_delimiter.delimiter, '/')
        self.assertEqual(len(out_delimiter.common_prefixes), 2)
        self.assertEqual(out_delimiter.common_prefixes[0].prefix, 'fun2/')
        self.assertEqual(out_delimiter.common_prefixes[1].prefix, 'fun3/')

    def test_list_part(self):
        bucket_name = self.bucket_name + "-test-list-part"
        key = self.random_key('.js')
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        out = self.client.create_multipart_upload(bucket_name, key).upload_id

    def test_complete_all(self):
        bucket_name = self.bucket_name + "-test-complete-all"
        key = self.random_key('.js')
        content = random_bytes(1024 * 1024 * 5)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        resp = self.client.create_multipart_upload(bucket=bucket_name, key=key)
        for i in range(2):
            self.client.upload_part(bucket=bucket_name, key=key, upload_id=resp.upload_id, part_number=i + 1,
                                    content=content)
        complete_out = self.client.complete_multipart_upload(bucket_name, key, resp.upload_id, complete_all=True)
        assert len(complete_out.complete_parts) == 2
        out = self.client.get_object(bucket_name, key)
        assert out.read() == content + content

    def test_multipart_callback(self):
        bucket_name = self.bucket_name + "-test-multipart-callback"
        self.bucket_delete.append(bucket_name)
        self.client.create_bucket(bucket_name)
        key = self.random_key('.js')
        mult_out = self.client.create_multipart_upload(bucket_name, key)
        parts = []
        content = random_bytes(5 * 1024 * 1024)
        for i in range(1, 3):
            upload_part_output = self.client.upload_part(bucket=bucket_name, key=key, upload_id=mult_out.upload_id,
                                                         part_number=i, content=content)
            parts.append(UploadedPart(i, upload_part_output.etag))

        callback = base64.b64encode(self.callback.encode('utf-8')).decode('utf-8')
        callback_var = base64.b64encode(self.callback_var.encode('utf-8')).decode('utf-8')
        complete_out = self.client.complete_multipart_upload(bucket_name, key, mult_out.upload_id, parts=parts,
                                                             callback=callback, callback_var=callback_var)

        self.assertEqual(complete_out.callback_result, '{"msg":"ok"}')
        self.assertTrue(len(complete_out.location) > 0)
        self.assertTrue(len(complete_out.etag) > 0)
        self.assertTrue(complete_out.hash_crc64_ecma > 0)

        get_out = self.client.get_object(bucket_name, key)
        self.assertEqual(get_out.read(), content + content)

        key = self.random_key('.js')
        mult_out = self.client.create_multipart_upload(bucket_name, key)
        parts = []
        content = random_bytes(5 * 1024 * 1024)
        for i in range(1, 3):
            upload_part_output = self.client.upload_part(bucket=bucket_name, key=key, upload_id=mult_out.upload_id,
                                                         part_number=i, content=content)
            parts.append(UploadedPart(i, upload_part_output.etag))

        callback_url = '{"callbackUrl" : "http://www.test.xxx.com"}'
        callback = base64.b64encode(callback_url.encode('utf-8')).decode('utf-8')
        with self.assertRaises(TosServerError) as cm:
            self.client.complete_multipart_upload(bucket_name, key, mult_out.upload_id, parts=parts,
                                                  callback=callback, callback_var=callback_var)
        self.assertEqual(cm.exception.status_code, 203)


if __name__ == '__main__':
    unittest.main()



import logging
import os
import threading
import time
from datetime import datetime, timedelta

import requests
from deprecated import deprecated
from tos.consts import ECS_DATE_FORMAT

from tos.exceptions import TosClientError

logger = logging.getLogger(__name__)

DEFAULT_FETCH_TIME = 5 * 60


class Credentials():
    def __init__(self, access_key_id, access_key_secret, security_token=None):
        self.access_key_id = access_key_id.strip()
        self.access_key_secret = access_key_secret.strip()
        self.security_token = security_token

    def get_ak(self):
        return self.access_key_id

    def get_sk(self):
        return self.access_key_secret

    def get_security_token(self):
        return self.security_token

    @deprecated(version='2.6.6', reason="please use get_ak")
    def get_access_key_id(self):
        return self.get_ak()

    @deprecated(version='2.6.6', reason="please use get_sk")
    def get_access_key_secret(self):
        return self.get_sk()


class CredentialsProvider():
    def get_credentials(self):
        return


class StaticCredentials(CredentialsProvider):
    """
    This class is deprecated and should not be used anymore.
    """
    @deprecated(version='2.6.6', reason="please use StaticCredentialsProvider")
    def __init__(self, access_key_id, access_key_secret, security_token=None):
        self.credentials = Credentials(access_key_id, access_key_secret, security_token)

    @deprecated(version='2.6.6', reason="please use StaticCredentialsProvider")
    def get_credentials(self):
        return self.credentials


class FederationToken():
    def __init__(self, access_key_id, access_key_secret, security_token, expiration, pre_fetch_sec=DEFAULT_FETCH_TIME):
        self.credential = Credentials(access_key_id, access_key_secret, security_token)
        self.expiration = expiration
        self.pre_fetch_sec = pre_fetch_sec

    def get_credentials(self):
        return self.credential

    def will_soon_expire(self):
        now = int(time.time())
        return now + self.pre_fetch_sec - self.expiration > 0

    def expire(self):
        return int(time.time()) > self.expiration


class FederationCredentials(CredentialsProvider):
    def __init__(self, get_credentials_func):
        self.get_credentials_func = get_credentials_func
        self.federationToken = None
        self.refreshing = 0
        self.__lock = threading.Lock()

    def get_credentials(self):
        # 不存在或者已经过期直接获取token
        if self.federationToken is None or self.federationToken.expire():
            return self._try_get_credential()
        # 快要过期且没有其他正在获取token的任务时，尝试去获取token
        if self.federationToken.will_soon_expire() and self.refreshing == 0:
            return self._try_get_credential()
        return self.federationToken.get_credentials()

    def _try_get_credential(self):
        with self.__lock:
            try:
                self.refreshing = 1
                # 再判断一次，因为可能已经被更新过了
                if self.federationToken is None or self.federationToken.will_soon_expire():
                    self.federationToken = self.get_credentials_func()
            except Exception as e:
                logger.error("get_credentials error: {0}".format(e))
                if self.federationToken is None:
                    raise
            finally:
                self.refreshing = 0
        return self.federationToken.get_credentials()


class StaticCredentialsProvider(CredentialsProvider):
    def __init__(self, access_key_id, access_key_secret, security_token=None):
        self.credentials = Credentials(access_key_id, access_key_secret, security_token)

    def get_credentials(self):
        return self.credentials


class EnvCredentialsProvider(CredentialsProvider):
    def __init__(self):
        access_key = os.environ.get('TOS_ACCESS_KEY')
        secret_key = os.environ.get('TOS_SECRET_KEY')
        security_token = os.environ.get('TOS_SECURITY_TOKEN')

        if access_key is None or secret_key is None:
            raise TosClientError('ak or sk is empty')

        self.credentials = Credentials(access_key, secret_key, security_token)

    def get_credentials(self):
        return self.credentials


class EcsCredentialsProvider(CredentialsProvider):
    ecs_url = 'http://100.96.0.96/volcstack/latest/iam/security_credentials/{}'

    def __init__(self, role_name):
        if role_name == '':
            raise TosClientError('ecs role name is empty')
        self._lock = threading.Lock()
        self.expires = None
        self.credentials = None
        self._ecs_url = EcsCredentialsProvider.ecs_url.format(role_name)

    def get_credentials(self):
        res = self._try_get_credentials()
        if res is not None:
            return res
        with self._lock:
            try:
                res = self._try_get_credentials()
                if res is not None:
                    return res

                res = requests.get(self._ecs_url, timeout=30)
                res_body = res.json()
                self.credentials = Credentials(res_body.get('AccessKeyId'), res_body.get('SecretAccessKey'),
                                               res_body.get('SessionToken'))
                self.expires = datetime.strptime(res_body.get('ExpiredTime'), ECS_DATE_FORMAT)
                return self.credentials
            except Exception as e:
                if self.expires is not None and datetime.now().timestamp() < self.expires.timestamp():
                    return self.credentials
                raise TosClientError('get ecs token failed', e)

    def _try_get_credentials(self):
        if self.expires is None or self.credentials is None:
            return None
        if datetime.now().timestamp() > (self.expires - timedelta(minutes=10)).timestamp():
            return None
        return self.credentials



# -*- coding: utf-8 -*-

import json
import unittest
from unittest import mock

from tests.common import MockResponse, TosTestCase


class TestBucket(TosTestCase):
    @mock.patch('requests.Session.request')
    def test_create_bucket(self, mock_request):
        mock_request.return_value = MockResponse(headers={'Location': self.location})

        res = self.client.create_bucket(self.bucket_name)
        self.assertEqual(res.location, self.location)

    @mock.patch('requests.Session.request')
    def test_head_bucket(self, mock_request):
        mock_request.return_value = MockResponse(headers={'x-tos-bucket-region': self.region})

        res = self.client.head_bucket(self.bucket_name)
        self.assertEqual(res.region, self.region)

    @mock.patch('requests.Session.request')
    def test_list_buckets(self, mock_request):
        mock_data = {
            'Owner': {
                'ID': 'id-test',
                'Name': 'name-test'
            },
            'Buckets': [
                {
                    'Name': 'bkt1',
                    'Location': self.location,
                    'CreationDate': self.date,
                    'ExtranetEndpoint': 'tos-ext.volces.com',
                    'IntranetEndpoint': 'tos-int.volces.com'
                },
                {
                    'Name': 'bkt2',
                    'Location': self.location,
                    'CreationDate': self.date,
                    'ExtranetEndpoint': 'tos-ext.volces.com',
                    'IntranetEndpoint': 'tos-int.volces.com'
                }
            ]
        }
        mock_request.return_value = MockResponse(body=json.dumps(mock_data))

        res = self.client.list_buckets()
        self.assertEqual(res.owner.id, 'id-test')
        self.assertEqual(res.owner.name, 'name-test')
        self.assertEqual(len(res.bucket_list), 2)
        self.assertEqual(res.bucket_list[0].name, 'bkt1')
        self.assertEqual(res.bucket_list[0].location, self.location)
        self.assertEqual(res.bucket_list[0].creation_date, self.date)
        self.assertEqual(res.bucket_list[0].extranet_endpoint, 'tos-ext.volces.com')
        self.assertEqual(res.bucket_list[0].intranet_endpoint, 'tos-int.volces.com')

        self.assertEqual(res.bucket_list[1].name, 'bkt2')
        self.assertEqual(res.bucket_list[1].location, self.location)
        self.assertEqual(res.bucket_list[1].creation_date, self.date)
        self.assertEqual(res.bucket_list[1].extranet_endpoint, 'tos-ext.volces.com')
        self.assertEqual(res.bucket_list[1].intranet_endpoint, 'tos-int.volces.com')

        mock_data = {
            'Owner': {
                'ID': 'id-test',
                'Name': 'name-test'
            }
        }
        mock_request.return_value = MockResponse(body=json.dumps(mock_data))
        res = self.client.list_buckets()
        self.assertEqual(res.owner.id, 'id-test')
        self.assertEqual(res.owner.name, 'name-test')
        self.assertEqual(len(res.bucket_list), 0)

    @mock.patch('requests.Session.request')
    def test_delete_bucket(self, mock_request):
        mock_request.return_value = MockResponse(status_code=204)

        res = self.client.delete_bucket(Bucket=self.bucket_name)
        self.assertEqual(res.status, 204)


if __name__ == '__main__':
    unittest.main()



# -*- coding: utf-8 -*-
import logging
import os
import tempfile
import time
import unittest
from functools import partial
from unittest.mock import patch

import tos
from tests.common import TosTestBase, random_bytes
from tos import DataTransferType
from tos.checkpoint import CancelHook, CheckPointStore
from tos.exceptions import TosClientError
from tos.utils import get_parent_directory_from_File


class NonlocalObject(object):
    def __init__(self, value):
        self.var = value


tos.set_logger(level=logging.INFO)


def upload_event(type, err, bucket, key, upload_id, file_path, checkpoint_file, part):
    print(type, err, bucket, key, upload_id, file_path, checkpoint_file, part)


def download_event(type, err, bucket, key, version_id, file_path, checkpint_file, tmp_file, download_part):
    print(type, err, bucket, key, version_id, file_path, checkpint_file, tmp_file, download_part)


def copy_event(type, err, bucket, key, upload_id, src_bucket, src_key, src_version_id, checkpoint_file,
               copy_part_info):
    print(type, err, bucket, key, upload_id, src_bucket, src_key, src_version_id, checkpoint_file, copy_part_info)


class TestUploadAndDownload(TosTestBase):
    def test_small_file(self):
        bucket_name = self.bucket_name + 'upload-download-small-file'
        self.__test_normal(bucket_name, 1023, upload_event=upload_event, download_event=download_event,
                           copy_event=copy_event)
        bucket_name = self.bucket_name + 'upload-download-small-file2'
        self.__test_normal(bucket_name, 1023)

    def test_large_file_with_single_threaded(self):
        bucket_name = self.bucket_name + 'upload-download-large-file-single-threaded'
        self.__test_normal(bucket_name, 10 * 1024 * 1024 + 3, thread_num=1, upload_event=upload_event,
                           download_event=download_event, copy_event=copy_event)
        bucket_name = self.bucket_name + 'upload-download-small-file2'
        self.__test_normal(bucket_name, 10 * 1024 * 1024 + 3, thread_num=1)

    def test_large_file_with_multi_threaded(self):
        bucket_name = self.bucket_name + 'upload-download-large-file-multi-threaded'
        self.__test_normal(bucket_name, 51 * 1024 * 1023, thread_num=7, upload_event=upload_event,
                           download_event=download_event, copy_event=copy_event)
        bucket_name = self.bucket_name + 'upload-download-small-file2'
        self.__test_normal(bucket_name, 51 * 1024 * 1023, thread_num=7)

    def test_threaded_larger_part_size(self):
        bucket_name = self.bucket_name + 'upload-threaded-larger-part-size'
        self.__test_normal(bucket_name, 50 * 1024 * 1023, thread_num=11)

    def test_download_fail_fist_part(self):
        bucket_name = self.bucket_name + 'download-fail-fist-part'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        key, content, filename_upload, filename_download = self.__prepare(40 * 1024 * 1025)
        self.__test_resume_upload(bucket_name, key, filename_upload, 40 * 1024 * 1025, content, [1])
        self.__test_resume_download(bucket_name, key, filename_download, 40 * 1024 * 1025, content, [1])

    def test_copy_part_fail_fist_part(self):
        bucket_name = self.bucket_name + 'download-fail-fist-part'
        src_bucket_name = self.bucket_name + 'download-fail-fist-part-src'
        self.client.create_bucket(bucket_name)
        self.client.create_bucket(src_bucket_name)
        self.bucket_delete.append(bucket_name)
        self.bucket_delete.append(src_bucket_name)
        key, content, filename_upload, filename_download = self.__prepare(40 * 1024 * 1025)
        self.client.put_object_from_file(src_bucket_name, key, filename_upload)
        self.__test_resume_copy(bucket_name, key, src_bucket_name, key, 40 * 1024 * 1025, content, [1])

    def test_copy_part_fail_last_part(self):
        bucket_name = self.bucket_name + 'download-fail-mid-part'
        src_bucket_name = self.bucket_name + 'download-fail-mid-part-src'
        self.client.create_bucket(bucket_name)
        self.client.create_bucket(src_bucket_name)
        self.bucket_delete.append(bucket_name)
        self.bucket_delete.append(src_bucket_name)
        key, content, filename_upload, filename_download = self.__prepare(40 * 1024 * 1025)
        self.client.put_object_from_file(src_bucket_name, key, filename_upload)
        self.__test_resume_copy(bucket_name, key, src_bucket_name, key, 40 * 1024 * 1025, content, [3])

    def test_copy_part_fail_mid_part(self):
        bucket_name = self.bucket_name + 'download-fail-mid-part'
        src_bucket_name = self.bucket_name + 'download-fail-mid-part-src'
        self.client.create_bucket(bucket_name)
        self.client.create_bucket(src_bucket_name)
        self.bucket_delete.append(bucket_name)
        self.bucket_delete.append(src_bucket_name)
        key, content, filename_upload, filename_download = self.__prepare(40 * 1024 * 1025)
        self.client.put_object_from_file(src_bucket_name, key, filename_upload)
        self.__test_resume_copy(bucket_name, key, src_bucket_name, key, 40 * 1024 * 1025, content, [2])

    def test_download_fail_last_part(self):
        bucket_name = self.bucket_name + 'download-fail-last-part'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        key, content, filename_upload, filename_download = self.__prepare(40 * 1024 * 1025)
        self.__test_resume_upload(bucket_name, key, filename_upload, 40 * 1024 * 1025, content, [3])
        self.__test_resume_download(bucket_name, key, filename_download, 40 * 1024 * 1025, content, [3])

    def test_download_fail_mid_part(self):
        bucket_name = self.bucket_name + 'download-fail-mid-part'
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        key, content, filename_upload, filename_download = self.__prepare(40 * 1024 * 1025)
        self.__test_resume_upload(bucket_name, key, filename_upload, 40 * 1024 * 1025, content, [2])
        self.__test_resume_download(bucket_name, key, filename_download, 40 * 1024 * 1025, content, [2])

    def test_download(self):
        bucket_name = self.bucket_name + 'download-dir'
        cwd = os.getcwd()
        key = self.random_key('.txt')
        content = random_bytes(1024 * 1024 * 5)
        self.client.create_bucket(bucket_name)
        self.bucket_delete.append(bucket_name)
        file_name = self.random_filename()
        with open(file_name, 'wb') as f:
            f.write(content)
        self.client.put_object_from_file(bucket_name, key, file_name)
        self.client.download_file(bucket_name, key, cwd + '/test.txt')
        self.assertFileContent(cwd + '/test.txt', content)

        content = content + b'1'
        self.client.put_object(bucket_name, key, content=content)
        self.client.download_file(bucket_name, key, cwd + '/test.txt')
        self.assertFileContent(cwd + '/test.txt', content)
        os.remove(cwd + '/test.txt')

        self.client.download_file(bucket_name, key, cwd + '/test/f1')
        self.assertFileContent(cwd + '/test/f1', content)

        self.client.download_file(bucket_name, key, cwd + '/dir1/')
        self.assertFileContent(cwd + '/dir1/' + key, content)
        os.remove(cwd + '/dir1/' + key)

        key = key + '/'
        self.client.put_object(bucket_name, key, content=content)
        self.client.download_file(bucket_name, key, cwd + '/dir3/')
        self.assertTrue(os.path.isdir(cwd + '/dir3/' + key))

    def __test_normal(self, bucket_name, file_size, part_size=5 * 1024 * 1024 + 1, thread_num=3, upload_event=None,
                      download_event=None, copy_event=None):
        def percentage(consumed_bytes, total_bytes, rw_once_bytes,
                       type: DataTransferType):
            if total_bytes:
                rate = int(100 * float(consumed_bytes) / float(total_bytes))
                print("rate:{}, consumed_bytes:{},total_bytes{}, rw_once_bytes:{}, type:{}".format(rate, consumed_bytes,
                                                                                                   total_bytes,
                                                                                                   rw_once_bytes, type))

        self.client.create_bucket(bucket=bucket_name)
        self.bucket_delete.append(bucket_name)
        key, content, filename_upload, filename_download = self.__prepare(file_size)
        self.client.upload_file(bucket=bucket_name, key=key, file_path=filename_upload, part_size=part_size,
                                task_num=thread_num, upload_event_listener=upload_event)

        self.client.download_file(bucket=bucket_name, key=key, file_path=filename_download, part_size=part_size,
                                  task_num=thread_num, download_event_listener=download_event)

        self.client.resumable_copy_object(bucket_name, 'key2', bucket_name, key, copy_event_listener=copy_event)

        self.assertFileContent(filename_upload, content)
        self.assertFileContent(filename_download, content)
        self.assertObjectContent(bucket_name, 'key2', content)

    def __test_resume_download(self, bucket_name, key, file_name, file_size, content, failed_parts):
        total = NonlocalObject(0)
        orig_download_part = tos.checkpoint._BreakpointDownloader._do_task

        def mock_download_part(self, part, failed_parts=None):
            if part.part_number in failed_parts:
                raise RuntimeError("Fail download_part for part: {0}".format(part.part_number))
            else:
                total.var += 1
                orig_download_part(self, part)

        self._test_resume(tos.checkpoint._BreakpointDownloader, '_do_task', mock_download_part, bucket_name, key,
                          file_name, file_size, content, failed_parts, 'download', self.client.download_file, total)

    def __test_resume_upload(self, bucket_name, key, file_name, file_size, content, failed_parts):
        total = NonlocalObject(0)
        orig_upload_part = tos.checkpoint._BreakpointUploader._do_task

        def mock_upload_part(self, part, failed_parts=None):
            if part.part_number in failed_parts:
                raise RuntimeError("Fail upload_part for part: {0}".format(part.part_number))
            else:
                total.var += 1
                orig_upload_part(self, part)

        self._test_resume(tos.checkpoint._BreakpointUploader, '_do_task', mock_upload_part, bucket_name, key,
                          file_name, file_size, content, failed_parts, 'upload', self.client.upload_file, total)

    def __test_resume_copy(self, bucket_name, key, src_bucket, src_key, file_size, content, failed_parts):
        total = NonlocalObject(0)
        orig_copy_part = tos.checkpoint._BreakpointResumableCopyObject._do_task

        def mock_copy_part(self, part, failed_parts=None):
            if part.part_number in failed_parts:
                raise RuntimeError("Fail copy for part: {0}".format(part.part_number))
            else:
                total.var += 1
                orig_copy_part(self, part)

        def percentage(copy_event_type, err, bucket, key, upload_id, src_bucket, src_key, src_version_id, checkpoint,
                       copy_part):
            print(copy_event_type, err, bucket, key, upload_id, src_bucket, src_key, src_version_id, checkpoint,
                  copy_part)

        with patch.object(tos.checkpoint._BreakpointResumableCopyObject, '_do_task',
                          side_effect=partial(mock_copy_part, failed_parts=failed_parts),
                          autospec=True):
            with self.assertRaises(TosClientError):
                self.client.resumable_copy_object(bucket_name, key, src_bucket, src_key, task_num=3,
                                                  copy_event_listener=percentage)

        store = CheckPointStore(tempfile.gettempdir(), '', 'copy')
        print('checkpoint', store.path(bucket_name, key, src_bucket, src_key))
        self.assertTrue(os.path.exists(store.path(bucket_name, key, src_bucket, src_key)))
        with patch.object(tos.checkpoint._BreakpointResumableCopyObject, '_do_task',
                          side_effect=partial(mock_copy_part, failed_parts=[]), autospec=True):
            self.client.resumable_copy_object(bucket_name, key, src_bucket, src_key, copy_event_listener=percentage,
                                              task_num=3)

        self.assertEqual(total.var, tos.utils.get_number(file_size, 20 * 1024 * 1024))
        self.assertFalse(os.path.exists(store.path(bucket_name, key, src_bucket=src_bucket, src_key=src_key)))
        self.assertObjectContent(bucket_name, key, content)

    def _test_resume(self, class_name, func_name, mock_func, bucket_name, key, file_name, file_size, content,
                     failed_parts, operation_type, api, total, src_bucket=None, src_key=None):
        bucket = bucket_name

        def percentage(consumed_bytes, total_bytes, rw_once_bytes,
                       type: DataTransferType):
            if total_bytes:
                rate = int(100 * float(consumed_bytes) / float(total_bytes))
                print("rate:{}, consumed_bytes:{},total_bytes{}, rw_once_bytes:{}, type:{}".format(rate, consumed_bytes,
                                                                                                   total_bytes,
                                                                                                   rw_once_bytes, type))

        with patch.object(class_name, func_name,
                          side_effect=partial(mock_func, failed_parts=failed_parts),
                          autospec=True):
            if src_bucket is None:
                self.assertRaises(TosClientError, api, bucket, key, file_name, data_transfer_listener=percentage,
                                  task_num=3)
            else:
                self.assertRaises(TosClientError, api, bucket, key, src_bucket, src_key, copy_event_listener=percentage,
                                  task_num=3)

        store = CheckPointStore(get_parent_directory_from_File(os.path.abspath(file_name)), file_name,
                                operation_type)
        if 1 not in failed_parts:
            self.assertTrue(os.path.exists(store.path(bucket, key)))
        if operation_type == 'download':
            self.assertTrue(os.path.exists(file_name + '.temp'))

        with patch.object(class_name, func_name,
                          side_effect=partial(mock_func, failed_parts=[]), autospec=True):
            if src_bucket is None:
                api(bucket, key, file_name, data_transfer_listener=percentage, task_num=3)
            else:
                api(bucket, key, src_bucket, src_key, copy_event_listener=percentage, task_num=3)

        self.assertEqual(total.var, tos.utils.get_number(file_size, 20 * 1024 * 1024))
        self.assertFalse(os.path.exists(file_name + '.temp'))
        self.assertFalse(os.path.exists(store.path(bucket, key)))
        self.assertFileContent(file_name, content)

    def __prepare(self, file_size, suffix=''):
        content = random_bytes(file_size)
        key = self.random_key(suffix)
        filename_upload = self.random_filename()
        filename_download = self.random_filename()
        with open(filename_upload, 'wb+') as f:
            f.write(content)
        return key, content, filename_upload, filename_download


if __name__ == '__main__':
    unittest.main()



import urllib.parse
from datetime import datetime

from requests.structures import CaseInsensitiveDict

from . import utils
from .enum import CannedType, GranteeType, PermissionType, StorageClassType, RedirectType, StatusType, \
    StorageClassInheritDirectiveType, VersioningStatusType, ProtocolType, CertStatus, AzRedundancyType, \
    convert_storage_class_type, convert_az_redundancy_type, convert_permission_type, convert_grantee_type, \
    convert_canned_type, convert_redirect_type, convert_status_type, convert_versioning_status_type, \
    convert_protocol_type, convert_cert_status, TierType, convert_tier_type
from .consts import CHUNK_SIZE
from .exceptions import TosClientError, make_server_error_with_exception
from .models import CommonPrefixInfo, DeleteMarkerInfo
from .utils import (get_etag, get_value, meta_header_decode,
                    parse_gmt_time_to_utc_datetime,
                    parse_modify_time_to_utc_datetime, _param_to_quoted_query, _make_virtual_host_url)


class ResponseInfo(object):
    def __init__(self, resp):
        self.request_id = resp.request_id
        self.id2 = get_value(resp.headers, "x-tos-id-2")
        self.status_code = resp.status
        self.header = resp.headers


class CreateBucketOutput(ResponseInfo):
    def __init__(self, resp):
        super(CreateBucketOutput, self).__init__(resp)
        self.location = get_value(self.header, "Location")


class HeadBucketOutput(ResponseInfo):
    def __init__(self, resp):
        super(HeadBucketOutput, self).__init__(resp)
        self.region = get_value(self.header, "x-tos-bucket-region")
        self.storage_class = get_value(self.header, "x-tos-storage-class", lambda x: convert_storage_class_type(x))
        self.az_redundancy = get_value(self.header, "x-tos-az-redundancy", lambda x: convert_az_redundancy_type(x))


class DeleteBucketOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketOutput, self).__init__(resp)


class ListedBucket(object):
    def __init__(self, name: str, location: str, creation_date: str, extranet_endpoint: str, intranet_endpoint: str):
        self.name = name
        self.location = location
        self.creation_date = creation_date
        self.extranet_endpoint = extranet_endpoint
        self.intranet_endpoint = intranet_endpoint

    def __str__(self):
        info = {'name': self.name, 'location': self.location, 'creation_date': self.creation_date,
                'extranet_endpoint': self.extranet_endpoint, 'intranet_endpoint': self.intranet_endpoint}

        return str(info)


class Owner(object):
    def __init__(self, id: str, display_name: str):
        self.id = id
        self.display_name = display_name

    def __str__(self):
        info = {'id': self.id, 'display_name': self.display_name}
        return str(info)


class ListBucketsOutput(ResponseInfo):
    def __init__(self, resp):
        super(ListBucketsOutput, self).__init__(resp)
        self.buckets = []  # ListedBucket
        self.owner = None  # Owner
        data = resp.json_read()
        self.owner = Owner(
            get_value(data['Owner'], 'ID'),
            get_value(data['Owner'], 'Name'),
        )

        bkt_list = get_value(data, 'Buckets') or []
        for bkt in bkt_list:
            self.buckets.append(ListedBucket(
                get_value(bkt, 'Name'),
                get_value(bkt, 'Location'),
                get_value(bkt, 'CreationDate'),
                get_value(bkt, 'ExtranetEndpoint'),
                get_value(bkt, 'IntranetEndpoint')))


class PutObjectOutput(ResponseInfo):
    def __init__(self, resp, callback=None):
        super(PutObjectOutput, self).__init__(resp)
        self.etag = get_etag(resp.headers)
        self.ssec_algorithm = get_value(resp.headers, "x-tos-server-side-encryption-customer-algorithm")
        self.ssec_key_md5 = get_value(resp.headers, "x-tos-server-side-encryption-customer-key-md5")
        self.version_id = get_value(resp.headers, "x-tos-version-id")
        self.hash_crc64_ecma = get_value(resp.headers, "x-tos-hash-crc64ecma", lambda x: int(x))
        if callback:
            self.callback_result = resp.read().decode("utf-8")


class CopyObjectOutput(ResponseInfo):
    def __init__(self, resp):
        super(CopyObjectOutput, self).__init__(resp)
        self.copy_source_version_id = get_value(resp.headers, "x-tos-copy-source-version-id")
        self.version_id = get_value(resp.headers, "x-tos-version-id")

        data = resp.json_read()
        self.etag = get_etag(data)
        self.last_modified = get_value(data, 'LastModified')
        if self.last_modified:
            self.last_modified = parse_modify_time_to_utc_datetime(self.last_modified)

        if not self.etag:
            make_server_error_with_exception(resp, data)


class DeleteObjectOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteObjectOutput, self).__init__(resp)
        self.version_id = get_value(resp.headers, 'x-tos-version-id')
        self.delete_marker = get_value(resp.headers, 'x-tos-delete-marker', bool)


class Deleted(object):
    def __init__(self, key: str, version_id: str = None, delete_marker=None,
                 delete_marker_version_id: str = None):
        self.key = key
        self.version_id = version_id
        self.delete_marker = delete_marker
        self.delete_marker_version_id = delete_marker_version_id


class ObjectTobeDeleted(object):
    def __init__(self, key: str, version_id: str = None):
        self.key = key
        self.version_id = version_id


class DeleteError(object):
    def __init__(self, key: str, version_id: str, code: str, message: str):
        self.key = key
        self.version_id = version_id
        self.code = code
        self.message = message


class DeleteObjectsOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteObjectsOutput, self).__init__(resp)
        self.deleted = []  # Deleted
        self.error = []  # DeleteError

        data = resp.json_read()

        delete_list = get_value(data, 'Deleted') or []
        for delete in delete_list:
            self.deleted.append(Deleted(
                get_value(delete, "Key"),
                get_value(delete, "VersionId"),
                get_value(delete, "DeleteMarker"),
                get_value(delete, "DeleteMarkerVersionId")
            ))
        err_list = get_value(data, 'Error') or []
        for err in err_list:
            self.error.append(DeleteError(
                get_value(err, "Key"),
                get_value(err, "VersionId"),
                get_value(err, "Code"),
                get_value(err, "Message"),
            ))


class Grantee(object):
    def __init__(self, type: GranteeType, id: str = None, display_name: str = None, canned: CannedType = None):
        self.type = type
        self.display_name = display_name
        self.id = id
        self.canned = canned

    def __str__(self):
        info = {'type': self.type, 'display_name': self.display_name, 'id': self.id, 'canned': self.canned}
        return str(info)


class Grant(object):
    def __init__(self, grantee: Grantee, permission: PermissionType):
        self.grantee = grantee
        self.permission = permission


class HeadObjectOutput(ResponseInfo):
    def __init__(self, resp):
        super(HeadObjectOutput, self).__init__(resp)
        self.etag = get_etag(resp.headers)
        self.version_id = get_value(resp.headers, "x-tos-version-id")
        self.sse_algorithm = get_value(resp.headers, "x-tos-server-side-encryption-customer-algorithm")
        self.sse_key_md5 = get_value(resp.headers, "x-tos-server-side-encryption-customer-key-MD5")
        self.website_redirect_location = get_value(resp.headers, "x-tos-website-redirect-location")
        self.hash_crc64_ecma = get_value(resp.headers, "x-tos-hash-crc64ecma", lambda x: int(x))
        self.storage_class = get_value(resp.headers, "x-tos-storage-class", lambda x: convert_storage_class_type(x))
        self.restore = get_value(resp.headers, "x-tos-restore")
        self.restore_expiry_days = get_value(resp.headers, "x-tos-restore-expiry-days", lambda x: int(x))
        self.restore_tier = get_value(resp.headers, "x-tos-restore-tier", lambda x: convert_tier_type(x))
        self.meta = CaseInsensitiveDict()
        self.object_type = get_value(resp.headers, "x-tos-object-type")
        if not self.object_type:
            self.object_type = "Normal"

        meta = {}
        for k in resp.headers:
            if k.startswith('x-tos-meta-'):
                meta[k[11:]] = resp.headers[k]
        self.meta = meta_header_decode(meta)

        self.last_modified = get_value(resp.headers, 'last-modified')
        if self.last_modified:
            self.last_modified = parse_gmt_time_to_utc_datetime(self.last_modified)
        self.expires = get_value(resp.headers, 'expires')
        if self.expires:
            self.expires = parse_gmt_time_to_utc_datetime(self.expires)

        if get_value(resp.headers, 'x-tos-delete-marker'):
            self.delete_marker = True
        else:
            self.delete_marker = False

        self.content_type = get_value(resp.headers, "content-type")
        self.content_length = get_value(resp.headers, "content-length", lambda x: int(x))
        self.cache_control = get_value(resp.headers, "cache-control")
        content_dis_str = get_value(resp.headers, 'content-disposition')
        if content_dis_str:
            self.content_disposition = urllib.parse.unquote(get_value(resp.headers, "content-disposition"))
        else:
            self.content_disposition = ''
        self.content_encoding = get_value(resp.headers, "content-encoding")
        self.content_language = get_value(resp.headers, "content-language")


class ListObjectsOutput(ResponseInfo):
    def __init__(self, resp):
        super(ListObjectsOutput, self).__init__(resp)
        data = resp.json_read()

        self.name = get_value(data, 'Name')
        self.prefix = get_value(data, 'Prefix')
        self.marker = get_value(data, 'Marker')
        self.max_keys = get_value(data, 'MaxKeys', int)
        self.next_marker = get_value(data, 'NextMarker')
        self.delimiter = get_value(data, 'Delimiter')
        self.contents = []  # ListedObject
        self.common_prefixes = []  # CommonPrefixInfo
        if get_value(data, 'EncodingType'):
            self.encoding_type = get_value(data, 'EncodingType')
        else:
            self.encoding_type = 'url'

        if get_value(data, 'IsTruncated'):
            self.is_truncated = get_value(data, 'IsTruncated', lambda x: bool(x))
        else:
            self.is_truncated = False

        common_prefix_list = get_value(data, 'CommonPrefixes') or []
        for common_prefix in common_prefix_list:
            self.common_prefixes.append(CommonPrefixInfo(get_value(common_prefix, 'Prefix')))

        object_list = get_value(data, 'Contents') or []
        for object in object_list:
            last_modified = get_value(object, 'LastModified')
            if last_modified:
                last_modified = parse_modify_time_to_utc_datetime(last_modified)
            object_info = ListedObject(
                get_value(object, 'Key'),
                last_modified=last_modified,
                etag=get_etag(object),
                size=get_value(object, 'Size', int),
                storage_class=get_value(object, 'StorageClass', lambda x: convert_storage_class_type(x)),
                hash_crc64_ecma=get_value(object, "HashCrc64ecma", lambda x: int(x))
            )
            owner_info = get_value(object, 'Owner')
            if owner_info:
                object_info.owner = Owner(
                    get_value(owner_info, "ID"),
                    get_value(owner_info, 'DisplayName')
                )
            self.contents.append(object_info)


class ListObjectsIterator(object):
    def __init__(self, req_func, max_key, bucket=None, key=None, method=None, data=None, headers=None, params=None,
                 func=None):
        self.req = req_func
        self.bucket = bucket
        self.key = key
        self.method = method
        self.data = data
        self.headers = headers
        self.params = params
        self.func = func
        self.max_key = max_key
        self.number = 0
        self.is_truncated = True

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        if self.is_truncated and self.new_max_key > 0:
            resp = self.req(bucket=self.bucket, method=self.method, key=self.key, data=self.data,
                            headers=self.headers, params=self.params)
            info = ListObjectType2Output(resp)
            self.is_truncated = info.is_truncated
            self.number += len(info.contents) + len(info.common_prefixes)
            self.params['max-keys'] = self.new_max_key
            self.params['continuation-token'] = info.next_continuation_token
            return info
        else:
            raise StopIteration

    @property
    def new_max_key(self):
        return self.max_key - self.number


class ListObjectType2Output(ResponseInfo):
    def __init__(self, resp):
        super(ListObjectType2Output, self).__init__(resp)
        data = resp.json_read()
        self.name = get_value(data, 'Name')
        self.prefix = get_value(data, 'Prefix')
        self.continuation_token = get_value(data, 'ContinuationToken')
        self.key_count = get_value(data, 'KeyCount', int)
        self.max_keys = get_value(data, 'MaxKeys', int)
        self.delimiter = get_value(data, 'Delimiter')
        self.common_prefixes = []  # CommonPrefixInfo
        self.contents = []  # ListedObject
        if get_value(data, 'IsTruncated'):
            self.is_truncated = get_value(data, 'IsTruncated', lambda x: bool(x))
        else:
            self.is_truncated = False

        if get_value(data, 'EncodingType'):
            self.encoding_type = get_value(data, 'EncodingType')
        else:
            self.encoding_type = 'url'

        self.next_continuation_token = get_value(data, 'NextContinuationToken')
        common_prefix_list = get_value(data, 'CommonPrefixes') or []
        for common_prefix in common_prefix_list:
            self.common_prefixes.append(CommonPrefixInfo(get_value(common_prefix, 'Prefix')))

        object_list = get_value(data, 'Contents') or []
        for object in object_list:
            last_modified = get_value(object, 'LastModified')
            if last_modified:
                last_modified = parse_modify_time_to_utc_datetime(last_modified)
            object_info = ListedObject(
                key=get_value(object, 'Key'),
                last_modified=last_modified,
                etag=get_etag(object),
                size=get_value(object, 'Size', int),
                storage_class=get_value(object, 'StorageClass', lambda x: convert_storage_class_type(x)),
                hash_crc64_ecma=get_value(object, "HashCrc64ecma", lambda x: int(x))
            )
            owner_info = get_value(object, 'Owner')
            if owner_info:
                object_info.owner = Owner(
                    get_value(owner_info, "ID"),
                    get_value(owner_info, 'DisplayName')
                )
            self.contents.append(object_info)

    def combine(self, listObjectType2Outputs: []):
        for output in listObjectType2Outputs:
            self.contents += output.contents
            for prefix in output.common_prefixes:
                self.common_prefixes.append(prefix)
        if len(listObjectType2Outputs) > 0:
            last = listObjectType2Outputs[len(listObjectType2Outputs) - 1]
            self.next_continuation_token = last.next_continuation_token
            self.is_truncated = last.is_truncated
            self.key_count = len(self.contents) + len(self.common_prefixes)

        return self


class ListedObject(object):
    def __init__(self, key: str, last_modified: datetime, etag: str, size: int, storage_class: StorageClassType,
                 hash_crc64_ecma: str, owner: Owner = None):
        self.key = key
        self.last_modified = last_modified
        self.etag = etag
        self.size = size
        self.owner = owner
        self.storage_class = storage_class
        self.hash_crc64_ecma = hash_crc64_ecma

    def __str__(self):
        info = {"key": self.key, "last_modified": self.last_modified, "etag": self.etag, "size": self.size, "owner": self.owner,
                "storage_class": self.storage_class, 'hash_crc64_ecma': self.hash_crc64_ecma}

        return str(info)


class ListedCommonPrefix(object):
    def __init__(self, prefix: str):
        self.prefix = prefix


class ListedObjectVersion(ListedObject):
    def __init__(self, key: str, last_modified: datetime, etag: str, size: int, storage_class: StorageClassType,
                 hash_crc64_ecma, owner: Owner = None, version_id: str = None, is_latest: bool = None):
        super(ListedObjectVersion, self).__init__(key, last_modified, etag, size, storage_class, hash_crc64_ecma, owner)
        self.version_id = version_id
        self.is_latest = is_latest


class ListObjectVersionsOutput(ResponseInfo):
    def __init__(self, resp):
        super(ListObjectVersionsOutput, self).__init__(resp)
        self.name = ''
        self.prefix = ''
        self.key_marker = ''
        self.version_id_marker = ''
        self.max_keys = None
        self.next_key_marker = ''
        self.next_version_id_marker = ''
        self.delimiter = ''
        self.is_truncated = False
        self.encoding_type = 'url'
        self.common_prefixes = []  # string list
        self.versions = []  # ListedObjectVersion list
        self.delete_markers = []  # DeleteMarkerInfo list

        data = resp.json_read()
        self.name = get_value(data, 'Name')
        self.prefix = get_value(data, 'Prefix')
        self.key_marker = get_value(data, 'KeyMarker')
        self.version_id_marker = get_value(data, 'VersionIdMarker')
        self.next_key_marker = get_value(data, 'NextKeyMarker')
        self.next_version_id_marker = get_value(data, 'NextVersionIdMarker')
        self.delimiter = get_value(data, 'Delimiter')
        self.max_keys = get_value(data, 'MaxKeys', int)

        if get_value(data, 'EncodingType'):
            self.encoding_type = get_value(data, 'EncodingType')
        else:
            self.encoding_type = 'url'

        if get_value(data, 'IsTruncated'):
            self.is_truncated = get_value(data, 'IsTruncated', lambda x: bool(x))
        else:
            self.is_truncated = False

        common_prefix_list = get_value(data, 'CommonPrefixes') or []
        for common_prefix in common_prefix_list:
            self.common_prefixes.append(CommonPrefixInfo(get_value(common_prefix, 'Prefix')))

        object_list = get_value(data, 'Versions') or []
        for object in object_list:
            last_modified = get_value(object, 'LastModified')
            if last_modified:
                last_modified = parse_modify_time_to_utc_datetime(last_modified)
            object_info = ListedObjectVersion(
                key=get_value(object, 'Key'),
                last_modified=last_modified,
                etag=get_etag(object),
                size=get_value(object, 'Size', lambda x: int(x)),
                storage_class=get_value(object, 'StorageClass', lambda x: convert_storage_class_type(x)),
                version_id=get_value(object, 'VersionId'),
                hash_crc64_ecma=get_value(object, "HashCrc64ecma", lambda x: int(x)),
                is_latest=get_value(object, "IsLatest", lambda x: bool(x))
            )
            owner_info = get_value(object, 'Owner')
            if owner_info:
                object_info.owner = Owner(
                    get_value(owner_info, "ID"),
                    get_value(owner_info, 'DisplayName')
                )
            self.versions.append(object_info)

        delete_marker_list = get_value(data, 'DeleteMarkers') or []
        for delete_marker in delete_marker_list:
            last_modified = get_value(delete_marker, 'LastModified')
            if last_modified:
                last_modified = parse_modify_time_to_utc_datetime(last_modified)
            delete_marker_info = DeleteMarkerInfo(
                get_value(delete_marker, 'Key'),
                get_value(delete_marker, 'IsLatest', lambda x: bool(x)),
                last_modified,
                get_value(delete_marker, 'VersionId')
            )
            owner_info = get_value(delete_marker, 'Owner')
            if owner_info:
                delete_marker_info.owner = Owner(
                    get_value(owner_info, "ID"),
                    get_value(owner_info, 'DisplayName')
                )
            self.delete_markers.append(delete_marker_info)


class Grants(object):
    def __init__(self, id: str, display_name: str, type: GranteeType, canned: CannedType):
        self.id = id
        self.display_name = display_name
        self.type = type
        self.canned = canned


class PutObjectACLOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutObjectACLOutput, self).__init__(resp)


class GetObjectACLOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetObjectACLOutput, self).__init__(resp)
        self.version_id = get_value(self.header, 'x-tos-version-id')
        self.owner = None
        self.grants = []
        data = resp.json_read()

        self.owner = Owner(
            get_value(data['Owner'], 'ID'),
            get_value(data['Owner'], 'DisplayName'),
        )

        grant_list = data['Grants'] or []
        for grant in grant_list:
            g = Grantee(
                id=get_value(grant['Grantee'], 'ID'),
                display_name=get_value(grant['Grantee'], 'DisplayName'),
                type=get_value(grant['Grantee'], 'Type', lambda x: convert_grantee_type(x)),
                canned=get_value(grant['Grantee'], 'Canned', lambda x: convert_canned_type(x)),
            )
            permission = get_value(grant, 'Permission', lambda x: convert_permission_type(x))
            self.grants.append(Grant(g, permission))


class SetObjectMetaOutput(ResponseInfo):
    def __init__(self, resp):
        super(SetObjectMetaOutput, self).__init__(resp)


class GetObjectOutput(HeadObjectOutput):
    def __init__(self, resp, progress_callback=None, rate_limiter=None, enable_crc=False, discard=0):
        super(GetObjectOutput, self).__init__(resp)
        self.enable_crc = enable_crc
        self.content_range = get_value(resp.headers, "content-range")
        self.content = resp
        if progress_callback:
            self.content = utils.add_progress_listener_func(data=resp, progress_callback=progress_callback,
                                                            size=self.content_length,
                                                            download_operator=True, is_response=True)
        if rate_limiter:
            self.content = utils.add_rate_limiter_func(data=self.content, rate_limiter=rate_limiter,
                                                       size=self.content_length,
                                                       is_response=True)

        if enable_crc:
            self.content = utils.add_crc_func(data=self.content, size=self.content_length, is_response=True)

    def read(self, amt=None):
        data = self.content.read(amt)
        if not data:
            if self.enable_crc and self.client_crc and self.content_range is None and self.hash_crc64_ecma is not None \
                    and self.client_crc != self.hash_crc64_ecma:
                raise TosClientError(
                    'client crc:{} is not equal to tos crc:{}'.format(self.client_crc, self.hash_crc64_ecma))
        return data

    @property
    def client_crc(self):
        if self.enable_crc:
            return self.content.crc
        return 0

    def __iter__(self):
        return self

    def __next__(self):
        return self.next()

    def next(self):
        content = self.read(CHUNK_SIZE)
        if content:
            return content
        else:
            if self.enable_crc and self.client_crc and self.content_range is None and self.client_crc != self.hash_crc64_ecma:
                raise TosClientError(
                    'client crc:{} is not equal to tos crc:{}'.format(self.client_crc, self.hash_crc64_ecma))
            raise StopIteration


class AppendObjectOutput(ResponseInfo):
    def __init__(self, resp):
        super(AppendObjectOutput, self).__init__(resp)
        self.version_id = get_value(resp.headers, "x-tos-version-id")
        self.next_append_offset = get_value(resp.headers, "x-tos-next-append-offset", lambda x: int(x))
        self.hash_crc64_ecma = get_value(resp.headers, "x-tos-hash-crc64ecma", lambda x: int(x))


class CreateMultipartUploadOutput(ResponseInfo):
    def __init__(self, resp):
        super(CreateMultipartUploadOutput, self).__init__(resp)

        self.ssec_algorithm = get_value(resp.headers, "x-tos-server-side-encryption-customer-algorithm")
        self.ssec_key_md5 = get_value(resp.headers, "x-tos-server-side-encryption-customer-key-md5")

        data = resp.json_read()
        self.bucket = get_value(data, "Bucket")
        self.key = get_value(data, "Key")
        self.upload_id = get_value(data, "UploadId")
        if get_value(data, 'EncodingType'):
            self.encoding_type = get_value(data, 'EncodingType')
        else:
            self.encoding_type = 'url'


class UploadPartOutput(ResponseInfo):
    def __init__(self, resp, number: int):
        super(UploadPartOutput, self).__init__(resp)
        self.part_number = number
        self.etag = get_etag(resp.headers)
        self.ssec_algorithm = get_value(resp.headers, 'x-tos-server-side-encryption-customer-algorithm')
        self.ssec_key_md5 = get_value(resp.headers, 'x-tos-server-side-encryption-customer-key-md5')
        self.hash_crc64_ecma = get_value(resp.headers, 'x-tos-hash-crc64ecma', lambda x: int(x))


class CompletePart(object):
    def __init__(self, etag, part_number):
        self.etag = etag
        self.part_number = part_number


class CompleteMultipartUploadOutput(ResponseInfo):
    def __init__(self, resp, callback: str = None):
        super(CompleteMultipartUploadOutput, self).__init__(resp)
        self.bucket = None
        self.key = None
        self.complete_parts = []
        self.callback_result = None
        self.version_id = get_value(resp.headers, 'x-tos-version-id')
        self.hash_crc64_ecma = get_value(resp.headers, 'x-tos-hash-crc64ecma', lambda x: int(x))
        if not callback:
            data = resp.json_read()
            self.bucket = get_value(data, 'Bucket')
            self.key = get_value(data, 'Key')
            self.etag = get_etag(data)
            self.location = get_value(data, 'Location')
            complete_part_info = get_value(data, 'CompletedParts') or []
            for part in complete_part_info:
                self.complete_parts.append(CompletePart(
                    etag=get_value(part, 'ETag'),
                    part_number=get_value(part, 'PartNumber')
                ))
        else:
            self.callback_result = resp.read().decode('utf-8')
            self.etag = get_etag(resp.headers)
            self.location = get_value(resp.headers, 'Location')


class AbortMultipartUpload(ResponseInfo):
    def __init__(self, resp):
        super(AbortMultipartUpload, self).__init__(resp)


class UploadPartCopyOutput(ResponseInfo):
    def __init__(self, resp, part_number):
        super(UploadPartCopyOutput, self).__init__(resp)
        data = resp.json_read()
        self.part_number = part_number
        self.etag = get_etag(data)
        self.last_modified = get_value(data, "LastModified")
        if self.last_modified:
            self.last_modified = parse_modify_time_to_utc_datetime(self.last_modified)
        self.copy_source_version_id = get_value(resp.headers, 'x-tos-copy-source-version-id')

        if not self.etag:
            raise make_server_error_with_exception(resp, data)


class ListedUpload(object):
    def __init__(self, key: str, upload_id: str, storage_class: StorageClassType, initiated: datetime,
                 owner: Owner = None):
        self.key = key
        self.upload_id = upload_id
        self.owner = owner
        self.storage_class = storage_class
        self.initiated = initiated


class ListMultipartUploadsOutput(ResponseInfo):
    def __init__(self, resp):
        super(ListMultipartUploadsOutput, self).__init__(resp)
        data = resp.json_read()

        self.bucket = get_value(data, 'Bucket')
        self.upload_id_marker = get_value(data, 'UploadIdMarker')
        self.next_key_marker = get_value(data, 'NextKeyMarker')
        self.next_upload_id_marker = get_value(data, 'NextUploadIdMarker')
        self.delimiter = get_value(data, 'Delimiter')
        self.prefix = get_value(data, 'Prefix')
        self.max_uploads = get_value(data, 'MaxUploads', lambda x: int(x))
        self.key_marker = get_value(data, 'KeyMarker')
        self.common_prefixes = []
        self.uploads = []
        if get_value(data, 'EncodingType'):
            self.encoding_type = get_value(data, 'EncodingType')
        else:
            self.encoding_type = 'url'

        if get_value(data, 'IsTruncated'):
            self.is_truncated = get_value(data, 'IsTruncated', lambda x: bool(x))
        else:
            self.is_truncated = False

        upload_list = get_value(data, 'Uploads') or []
        for upload in upload_list:
            initiated = get_value(upload, 'Initiated')
            if initiated:
                initiated = parse_modify_time_to_utc_datetime(initiated)
            multipart_upload_info = ListedUpload(
                get_value(upload, 'Key'),
                get_value(upload, 'UploadId'),
                get_value(upload, 'StorageClass', lambda x: convert_storage_class_type(x)),
                initiated,
            )

            owner = get_value(upload, 'Owner')
            if owner:
                id = get_value(owner, 'ID')
                name = get_value(owner, 'DisplayName')
                multipart_upload_info.owner = Owner(id, name)

            self.uploads.append(multipart_upload_info)

        common_prefix_list = get_value(data, 'CommonPrefixes') or []
        for common_prefix in common_prefix_list:
            self.common_prefixes.append(CommonPrefixInfo(get_value(common_prefix, 'Prefix')))


class ListPartsOutput(ResponseInfo):
    def __init__(self, resp):
        super(ListPartsOutput, self).__init__(resp)

        data = resp.json_read()

        self.bucket = get_value(data, 'Bucket')
        self.key = get_value(data, 'Key')
        self.upload_id = get_value(data, 'UploadId')
        self.part_number_marker = get_value(data, 'PartNumberMarker', lambda x: int(x))
        self.next_part_number_marker = get_value(data, 'NextPartNumberMarker', int)
        self.max_parts = get_value(data, 'MaxParts', lambda x: int(x))
        self.storage_class = get_value(data, 'StorageClass', lambda x: convert_storage_class_type(x))
        self.parts = []
        if get_value(data, 'EncodingType'):
            self.encoding_type = get_value(data, 'EncodingType')
        else:
            self.encoding_type = 'url'

        if get_value(data, 'IsTruncated'):
            self.is_truncated = get_value(data, 'IsTruncated', lambda x: bool(x))
        else:
            self.is_truncated = False

        owner = get_value(data, 'Owner')
        if owner:
            id = get_value(owner, 'ID')
            name = get_value(owner, 'DisplayName')
            self.owner = Owner(id, name)

        parts = get_value(data, 'Parts') or []
        for part in parts:
            last_modified = get_value(part, 'LastModified')
            if last_modified:
                last_modified = parse_modify_time_to_utc_datetime(last_modified)
            self.parts.append(UploadedPart(
                part_number=get_value(part, 'PartNumber'),
                last_modified=last_modified,
                etag=get_etag(part),
                size=get_value(part, 'Size', int)
            ))


class PutBucketCorsOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketCorsOutput, self).__init__(resp)


class GetBucketCorsOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketCorsOutput, self).__init__(resp)
        data = resp.json_read()
        self.cors_rules = []
        cors_rules = get_value(data, 'CORSRules') or []
        for rule in cors_rules:
            self.cors_rules.append(CORSRule(
                allowed_origins=get_value(rule, 'AllowedOrigins'),
                allowed_methods=get_value(rule, 'AllowedMethods'),
                allowed_headers=get_value(rule, 'AllowedHeaders'),
                expose_headers=get_value(rule, 'ExposeHeaders'),
                max_age_seconds=get_value(rule, 'MaxAgeSeconds', lambda x: int(x))
            ))


class DeleteBucketCorsOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketCorsOutput, self).__init__(resp)


class Condition(object):
    def __init__(self, http_code: int = None, key_prefix: str = None, key_suffix: str = None):
        self.http_code = http_code
        self.key_prefix = key_prefix
        self.key_suffix = key_suffix


class ReplaceKeyPrefix(object):
    def __init__(self, key_prefix: str = None, replace_with: str = None):
        self.key_prefix = key_prefix
        self.replace_with = replace_with


class Transform(object):
    def __init__(self, with_key_prefix: str = None, with_key_suffix: str = None,
                 replace_key_prefix: ReplaceKeyPrefix = None):
        self.with_key_prefix = with_key_prefix
        self.with_key_suffix = with_key_suffix
        self.replace_key_prefix = replace_key_prefix


class SourceEndpoint(object):
    def __init__(self, primary: [] = None, follower: [] = None):
        self.primary = primary
        self.follower = follower


class PublicSource(object):
    def __init__(self, source_endpoint: SourceEndpoint = None, fixed_endpoint: bool = None):
        self.source_endpoint = source_endpoint
        self.fixed_endpoint = fixed_endpoint


class MirrorHeader(object):
    def __init__(self, pass_all: bool = None, pass_headers: [] = None, remove: [] = None):
        self.pass_all = pass_all
        self.pass_headers = pass_headers
        self.remove = remove


class Redirect(object):
    def __init__(self, redirect_type: RedirectType = None, public_source: PublicSource = None,
                 fetch_source_on_redirect: bool = None, pass_query: bool = None, follow_redirect: bool = None,
                 mirror_header: MirrorHeader = None, transform: Transform = None):
        self.redirect_type = redirect_type
        self.fetch_source_on_redirect = fetch_source_on_redirect
        self.public_source = public_source
        self.pass_query = pass_query
        self.follow_redirect = follow_redirect
        self.mirror_header = mirror_header
        self.transform = transform


class Rule(object):
    def __init__(self, id: str = None, condition: Condition = None, redirect: Redirect = None):
        self.id = id
        self.condition = condition
        self.redirect = redirect


class PutBucketMirrorBackOutPut(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketMirrorBackOutPut, self).__init__(resp)


# class GetBucketMirrorBackOutPut(ResponseInfo):
#     def __init__(self, resp):
#         super(GetBucketMirrorBackOutPut, self).__init__(resp)
#         data = resp.json_read()
#         self.cors_rules = []
#         rules = get_value(data, 'Rules')
#         for rule in rules:
#             self.cors_rules.append(Rule(
#                 id=get_value(rule, 'ID', lambda x: int(x)),
#                 condition=Condition(get_value(rule, 'HttpCode', lambda x: int(x)), get_value(rule, 'ObjectKeyPrefix')),
#                 redirect=Redirect()
#             ))


class DeleteBucketMirrorBackOutPut(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketMirrorBackOutPut, self).__init__(resp)


class UploadedPart(object):
    def __init__(self, part_number: int, etag: int, size: int = None, last_modified: datetime = None):
        self.part_number = part_number
        self.etag = etag
        if size:
            self.size = size
        if last_modified:
            self.last_modified = last_modified


class PreSignedURLOutput(object):
    def __init__(self, signed_url, signed_header):
        self.signed_url = signed_url
        self.signed_header = signed_header


class UploadFileOutput(object):
    def __init__(self, resp: CompleteMultipartUploadOutput, ssec_algorithm, ssec_key_md5, upload_id, encoding_type):
        self.request_id = resp.request_id
        self.id2 = resp.id2
        self.status_code = resp.status_code
        self.header = resp.header
        self.bucket = resp.bucket
        self.key = resp.key
        self.etag = resp.etag
        self.location = resp.location
        self.upload_id = upload_id
        self.version_id = resp.version_id
        self.hash_crc64_ecma = resp.hash_crc64_ecma
        self.encoding_type = encoding_type
        self.ssec_algorithm = ssec_algorithm
        self.ssec_key_md5 = ssec_key_md5


class PartInfo(object):
    def __init__(self, part_number, part_size, offset, etag, hash_crc64_ecma, is_completed):
        self.part_number = part_number
        self.part_size = part_size
        self.offset = offset
        self.etag = etag
        self.hash_crc64_ecma = hash_crc64_ecma
        self.is_completed = is_completed

    def __str__(self):
        info = {'part_number': self.part_number, 'part_size': self.part_size, 'offset': self.offset, 'etag': self.etag,
                'hash_crc64_ecma': self.hash_crc64_ecma, 'is_completed': self.is_completed}
        return str(info)


class DownloadPartInfo(object):
    def __init__(self, part_number, range_start, range_end, hash_crc64_ecma, is_completed):
        self.part_number = part_number
        self.range_start = range_start
        self.range_end = range_end
        self.hash_crc64_ecma = hash_crc64_ecma
        self.is_completed = is_completed

    @property
    def size(self):
        return self.range_end - self.range_start


class _PartToDo(object):
    def __init__(self, part_number, start, end, part_crc=None):
        self.part_number = part_number
        self.start = start
        self.end = end
        self.part_crc = part_crc

    @property
    def size(self):
        return self.end - self.start

    def __hash__(self):
        return hash(self.__key)

    def __eq__(self, other):
        return self.__key == other.__key

    @property
    def __key(self):
        return self.part_number, self.start, self.end

    def __str__(self):
        info = {'part_number': self.part_number, 'start': self.start, 'end': self.end, 'part_crc': self.part_crc}
        return str(info)


class CORSRule(object):
    def __init__(self, allowed_origins: [] = None, allowed_methods: [] = None, allowed_headers: [] = None,
                 expose_headers: [] = None,
                 max_age_seconds: int = None):
        self.allowed_origins = allowed_origins
        self.allowed_methods = allowed_methods
        self.allowed_headers = allowed_headers
        self.expose_headers = expose_headers
        self.max_age_seconds = max_age_seconds


class Tag(object):
    def __init__(self, key: str = None, value: str = None):
        self.key = key
        self.value = value


class PutBucketStorageClassOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketStorageClassOutput, self).__init__(resp)


class GetBucketLocationOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketLocationOutput, self).__init__(resp)
        data = resp.json_read()
        self.region = get_value(data, 'Region')
        self.extranet_endpoint = get_value(data, 'ExtranetEndpoint')
        self.intranet_endpoint = get_value(data, 'IntranetEndpoint')


class BucketLifeCycleExpiration(object):
    def __init__(self, days: int = None, date: datetime = None):
        self.days = days
        self.date = date


class BucketLifeCycleNoCurrentVersionExpiration(object):
    def __init__(self, no_current_days: int = None):
        self.no_current_days = no_current_days


class BucketLifeCycleAbortInCompleteMultipartUpload(object):
    def __init__(self, days_after_init: int = None):
        self.days_after_init = days_after_init


class BucketLifeCycleTransition(object):
    def __init__(self, storage_class: StorageClassType = None, days: int = None, date: datetime = None):
        self.storage_class = storage_class
        self.days = days
        self.date = date


class BucketLifeCycleNonCurrentVersionTransition(object):
    def __init__(self, storage_class: StorageClassType = None, non_current_days: int = None):
        self.storage_class = storage_class
        self.non_current_days = non_current_days


class BucketLifeCycleRule(object):

    def __init__(self, status: StatusType = None,
                 expiration: BucketLifeCycleExpiration = None,
                 no_current_version_expiration: BucketLifeCycleNoCurrentVersionExpiration = None,
                 abort_in_complete_multipart_upload: BucketLifeCycleAbortInCompleteMultipartUpload = None,
                 tags: [] = None,
                 transitions: [] = None,
                 non_current_version_transitions: [] = None,
                 id: str = None,
                 prefix: str = None):
        self.id = id
        self.prefix = prefix
        self.status = status
        self.expiration = expiration
        self.no_current_version_expiration = no_current_version_expiration
        self.abort_in_complete_multipart_upload = abort_in_complete_multipart_upload
        self.tags = tags
        self.transitions = transitions
        self.non_current_version_transitions = non_current_version_transitions


class PutBucketLifecycleOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketLifecycleOutput, self).__init__(resp)


class GetBucketLifecycleOutput(ResponseInfo):
    def __init__(self, resp):
        self.rules = []
        super(GetBucketLifecycleOutput, self).__init__(resp)
        data = resp.json_read()
        rules_json = get_value(data, 'Rules') or []
        for rule_json in rules_json:
            rule = BucketLifeCycleRule()
            rule.id = get_value(rule_json, 'ID')
            rule.prefix = get_value(rule_json, 'Prefix')
            if get_value(rule_json, 'Status'):
                rule.status = get_value(rule_json, 'Status', lambda x: convert_status_type(x))

            expiration_json = get_value(rule_json, 'Expiration')
            non_current_version_expiration_json = get_value(rule_json, 'NoncurrentVersionExpiration')
            abort_incomplete_multipart_upload_json = get_value(rule_json, 'AbortIncompleteMultipartUpload')
            tags_json = get_value(rule_json, 'Tags') or []
            transitions_json = get_value(rule_json, 'Transitions') or []
            non_current_version_transitions_json = get_value(rule_json, 'NoncurrentVersionTransitions') or []

            if expiration_json:
                bucket_expiration = BucketLifeCycleExpiration()
                bucket_expiration.days = get_value(expiration_json, 'Days', int)
                if get_value(expiration_json, 'Date'):
                    bucket_expiration.date = parse_modify_time_to_utc_datetime(get_value(expiration_json, 'Date'))
                rule.expiration = bucket_expiration

            if non_current_version_transitions_json:
                rule.non_current_version_transitions = []
                for vt in non_current_version_transitions_json:
                    tr = BucketLifeCycleNonCurrentVersionTransition()
                    tr.storage_class = get_value(vt, 'StorageClass', lambda x: convert_storage_class_type(x))
                    tr.non_current_days = get_value(vt, 'NoncurrentDays', int)
                    rule.non_current_version_transitions.append(tr)

            if transitions_json:
                rule.transitions = []
                for transition_json in transitions_json:
                    ts = BucketLifeCycleTransition()
                    ts.storage_class = get_value(transition_json, 'StorageClass',
                                                 lambda x: convert_storage_class_type(x))
                    ts.days = get_value(transition_json, 'Days', int)
                    if get_value(transition_json, 'Date'):
                        ts.date = parse_modify_time_to_utc_datetime(get_value(transition_json, 'Date'))
                    rule.transitions.append(ts)

            if tags_json:
                rule.tags = []
                for tag_json in tags_json:
                    tag = Tag()
                    tag.key = get_value(tag_json, 'Key')
                    tag.value = get_value(tag_json, 'Value')
                    rule.tags.append(tag)

            if abort_incomplete_multipart_upload_json:
                abort = BucketLifeCycleAbortInCompleteMultipartUpload()
                abort.days_after_init = get_value(abort_incomplete_multipart_upload_json, 'DaysAfterInitiation',
                                                  int)
                rule.abort_in_complete_multipart_upload = abort

            if non_current_version_expiration_json:
                exp = BucketLifeCycleNoCurrentVersionExpiration()
                exp.no_current_days = get_value(non_current_version_expiration_json, 'NoncurrentDays', int)
                rule.no_current_version_expiration = exp
            self.rules.append(rule)


class DeleteBucketLifecycleOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketLifecycleOutput, self).__init__(resp)


class PutBucketPolicyOutPut(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketPolicyOutPut, self).__init__(resp)


class GetBucketPolicyOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketPolicyOutput, self).__init__(resp)
        self.policy = resp.read().decode("utf-8")


class DeleteBucketPolicy(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketPolicy, self).__init__(resp)


class GetBucketMirrorBackOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketMirrorBackOutput, self).__init__(resp)
        self.rules = []
        data = resp.json_read()
        get_rules = get_value(data, 'Rules') or []
        for rule in get_rules:
            id = get_value(rule, 'ID')
            cond = get_value(rule, 'Condition')
            condition = None
            red = get_value(rule, 'Redirect')
            redirect = None
            if cond:
                condition = Condition(
                    http_code=get_value(cond, 'HttpCode', int),
                    key_prefix=get_value(cond, 'KeyPrefix', str),
                    key_suffix=get_value(cond, 'KeySuffix', str)
                )
            if red:
                redirect = Redirect()
                redirect.redirect_type = get_value(red, 'RedirectType', lambda x: convert_redirect_type(x))
                redirect.fetch_source_on_redirect = get_value(red, 'FetchSourceOnRedirect', lambda x: bool(x))

                if get_value(red, 'PublicSource'):
                    redirect.public_source = PublicSource(
                        fixed_endpoint=get_value(get_value(red, 'PublicSource'), 'FixedEndpoint', lambda x: bool(x)))
                    if get_value(get_value(red, 'PublicSource'), 'SourceEndpoint'):
                        redirect.public_source.source_endpoint = SourceEndpoint(
                            primary=get_value(get_value(get_value(red, 'PublicSource'), 'SourceEndpoint'), 'Primary'),
                            follower=get_value(get_value(get_value(red, 'PublicSource'), 'SourceEndpoint'), 'Follower')
                        )
                redirect.pass_query = get_value(red, 'PassQuery', lambda x: bool(x))
                redirect.follow_redirect = get_value(red, 'FollowRedirect', lambda x: bool(x))
                if get_value(red, 'MirrorHeader'):
                    redirect.mirror_header = MirrorHeader(
                        pass_all=get_value(get_value(red, 'MirrorHeader'), 'PassAll', lambda x: bool(x)),
                        pass_headers=get_value(get_value(red, 'MirrorHeader'), 'Pass'),
                        remove=get_value(get_value(red, 'MirrorHeader'), 'Remove')
                    )
                if get_value(red, 'Transform'):
                    redirect.transform = Transform(
                        with_key_prefix=get_value(get_value(red, 'Transform'), 'WithKeyPrefix'),
                        with_key_suffix=get_value(get_value(red, 'Transform'), 'WithKeySuffix'),
                    )
                    if get_value(get_value(red, 'Transform'), 'ReplaceKeyPrefix'):
                        redirect.transform.replace_key_prefix = ReplaceKeyPrefix(
                            key_prefix=get_value(get_value(get_value(red, 'Transform'), 'ReplaceKeyPrefix'),
                                                 'KeyPrefix'),
                            replace_with=get_value(get_value(get_value(red, 'Transform'), 'ReplaceKeyPrefix'),
                                                   'ReplaceWith')
                        )
                r = Rule(id=id, condition=condition, redirect=redirect)
                self.rules.append(r)


class PutObjectTaggingOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutObjectTaggingOutput, self).__init__(resp)
        self.version_id = get_value(resp.headers, 'x-tos-version-id')


class GetObjectTaggingOutPut(ResponseInfo):
    def __init__(self, resp):
        super(GetObjectTaggingOutPut, self).__init__(resp)
        self.version_id = self.version_id = get_value(resp.headers, 'x-tos-version-Id')
        self.tag_set = []
        data = resp.json_read()
        tag_set = get_value(data, 'TagSet')
        if tag_set:
            tags = get_value(tag_set, 'Tags')
            if tags:
                for tag in tags:
                    self.tag_set.append(Tag(
                        get_value(tag, 'Key'),
                        get_value(tag, 'Value')
                    ))


class DeleteObjectTaggingOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteObjectTaggingOutput, self).__init__(resp)
        self.version_id = get_value(resp.headers, 'x-tos-version-id')


class DeleteBucketMirrorBackOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketMirrorBackOutput, self).__init__(resp)


class PutBucketACLOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketACLOutput, self).__init__(resp)


class GetBucketACLOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketACLOutput, self).__init__(resp)
        self.owner = None
        self.grants = []
        data = resp.json_read()

        self.owner = Owner(
            get_value(data['Owner'], 'ID'),
            get_value(data['Owner'], 'DisplayName'),
        )

        grant_list = data['Grants'] or []
        for grant in grant_list:
            g = Grantee(
                id=get_value(grant['Grantee'], 'ID'),
                display_name=get_value(grant['Grantee'], 'DisplayName'),
                type=get_value(grant['Grantee'], 'Type', lambda x: convert_grantee_type(x)),
                canned=get_value(grant['Grantee'], 'Canned', lambda x: convert_canned_type(x)),
            )
            permission = get_value(grant, 'Permission', lambda x: convert_permission_type(x))
            self.grants.append(Grant(g, permission))


class PreSignedPostSignatureOutPut(object):
    def __init__(self):
        self.origin_policy = None  # 签名前 policy
        self.policy = None  # 签名后policy 用于表单域policy
        self.algorithm = None  # 用于表单域 x-tos-algorithm
        self.credential = None  # 用于表单域 x-tos-credential
        self.date = None  # 用于表单域 x-tos-date
        self.signature = None  # 用于表单域 x-tos-signature


class PostSignatureCondition(object):
    def __init__(self, key: str, value: str, operator=None):
        self.key = key
        self.value = value
        self.operator = operator


class PolicySignatureCondition(object):
    def __init__(self, key: str, value: str, operator=None):
        self.key = key
        self.value = value
        self.operator = operator


class ContentLengthRange(object):
    def __init__(self, range_start: int, range_end: int):
        self.start = range_start
        self.end = range_end


class FetchObjectOutput(ResponseInfo):
    def __init__(self, resp):
        super(FetchObjectOutput, self).__init__(resp)
        self.version_id = get_value(resp.headers, 'x-version-id')
        self.ssec_algorithm = get_value(resp.headers, 'x-server-side-encryption-customer-key-md5')
        self.ssec_key_md5 = get_value(resp.headers, 'x-server-side-encryption-customer-algorithm')
        data = resp.json_read()
        self.etag = get_etag(data)


class PutFetchTaskOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutFetchTaskOutput, self).__init__(resp)
        data = resp.json_read()
        self.task_id = get_value(data, 'TaskId')


class PutBucketReplicationOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketReplicationOutput, self).__init__(resp)


class Destination(object):
    def __init__(self, bucket: str = None, location: str = None, storage_class: StorageClassType = None,
                 storage_class_inherit_directive: StorageClassInheritDirectiveType = None):
        self.bucket = bucket
        self.location = location
        self.storage_class = storage_class
        self.storage_class_inherit_directive = storage_class_inherit_directive


class Progress(object):
    def __init__(self, historical_object: float, new_object: str):
        self.historical_object = historical_object
        self.new_object = new_object


class ReplicationRule(object):
    def __init__(self, id: str = None, status: StatusType = None, prefix_set: [] = None,
                 destination: Destination = None,
                 historical_object_replication: StatusType = None, progress: Progress = None):
        self.id = id
        self.status = status
        self.prefix_set = prefix_set
        self.destination = destination
        self.historical_object_replication = historical_object_replication
        self.progress = progress


class GetBucketReplicationOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketReplicationOutput, self).__init__(resp)
        self.rules = []
        data = resp.json_read()
        rules_json = get_value(data, 'Rules') or []
        for rule_json in rules_json:
            replication_rule = ReplicationRule()
            replication_rule.id = get_value(rule_json, 'ID')
            replication_rule.status = get_value(rule_json, 'Status', lambda x: convert_status_type(x))
            replication_rule.prefix_set = get_value(rule_json, 'PrefixSet')
            replication_rule.historical_object_replication = get_value(rule_json, 'HistoricalObjectReplication',
                                                                       lambda x: convert_status_type(x))

            destination_json = get_value(rule_json, 'Destination')
            progress_json = get_value(rule_json, 'Progress')
            if destination_json:
                replication_rule.destination = Destination(
                    bucket=get_value(destination_json, 'Bucket'),
                    location=get_value(destination_json, 'Location'),
                    storage_class=get_value(destination_json, 'StorageClass', lambda x: convert_storage_class_type(x)),
                    storage_class_inherit_directive=get_value(destination_json, 'StorageClassInheritDirective',
                                                              lambda x: StorageClassInheritDirectiveType(x)))

            if progress_json:
                replication_rule.progress = Progress(
                    historical_object=get_value(progress_json, 'HistoricalObject', lambda x: float(x)),
                    new_object=get_value(progress_json, 'NewObject'))
            self.rules.append(replication_rule)


class DeleteBucketReplicationOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketReplicationOutput, self).__init__(resp)


class PutBucketVersioningOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketVersioningOutput, self).__init__(resp)


class GetBucketVersionOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketVersionOutput, self).__init__(resp)
        data = resp.json_read()
        self.status = get_value(data, 'Status', lambda x: convert_versioning_status_type(x))


class RedirectAllRequestsTo(object):
    def __init__(self, host_name, protocol):
        self.host_name = host_name
        self.protocol = protocol


class IndexDocument(object):
    def __init__(self, suffix: str = None, forbidden_sub_dir: bool = None):
        self.suffix = suffix
        self.forbidden_sub_dir = forbidden_sub_dir


class ErrorDocument(object):
    def __init__(self, key: str):
        self.key = key


class RoutingRuleCondition(object):
    def __init__(self, key_prefix_equals=None, http_error_code_returned_equals=None):
        self.key_prefix_equals = key_prefix_equals
        self.http_error_code_returned_equals = http_error_code_returned_equals


class RoutingRuleRedirect(object):
    def __init__(self, protocol: ProtocolType = None, host_name: str = None, replace_key_prefix_with: str = None,
                 replace_key_with: str = None,
                 http_redirect_code: int = None):
        self.protocol = protocol
        self.host_name = host_name
        self.replace_key_prefix_with = replace_key_prefix_with
        self.replace_key_with = replace_key_with
        self.http_redirect_code = http_redirect_code


class RoutingRules(object):
    def __init__(self, rules: []):
        self.rules = rules


class RoutingRule(object):
    def __init__(self, condition: RoutingRuleCondition = None, redirect: RoutingRuleRedirect = None):
        self.condition = condition
        self.redirect = redirect


class PutBucketWebsiteOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketWebsiteOutput, self).__init__(resp)


class GetBucketWebsiteOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketWebsiteOutput, self).__init__(resp)
        self.redirect_all_requests_to = None
        self.index_document = None
        self.error_document = None
        self.routing_rules = []
        data = resp.json_read()
        redirect_all_requests_to = get_value(data, 'RedirectAllRequestsTo')
        index_document = get_value(data, 'IndexDocument')
        error_document = get_value(data, 'ErrorDocument')
        routing_rules = get_value(data, 'RoutingRules') or []
        if redirect_all_requests_to:
            self.redirect_all_requests_to = RedirectAllRequestsTo(
                host_name=get_value(redirect_all_requests_to, 'HostName'),
                protocol=get_value(redirect_all_requests_to, 'Protocol'))
        if index_document:
            self.index_document = IndexDocument(suffix=get_value(index_document, 'Suffix'),
                                                forbidden_sub_dir=get_value(index_document, 'ForbiddenSubDir'))

        if error_document:
            self.error_document = ErrorDocument(key=get_value(error_document, 'Key'))

        for rule_json in routing_rules:
            rule = RoutingRule()
            condition = get_value(rule_json, 'Condition')
            redirect = get_value(rule_json, 'Redirect')
            if condition:
                rule.condition = RoutingRuleCondition(
                    http_error_code_returned_equals=get_value(condition, 'HttpErrorCodeReturnedEquals'),
                    key_prefix_equals=get_value(condition, 'KeyPrefixEquals'))
            if redirect:
                rule.redirect = RoutingRuleRedirect(
                    protocol=get_value(redirect, 'Protocol', lambda x: convert_protocol_type(x)),
                    host_name=get_value(redirect, 'HostName'),
                    replace_key_prefix_with=get_value(redirect, 'ReplaceKeyPrefixWith'),
                    replace_key_with=get_value(redirect, 'ReplaceKeyWith'),
                    http_redirect_code=get_value(redirect, 'HttpRedirectCode'))
            self.routing_rules.append(rule)


class FilterRule(object):
    def __init__(self, name: str = None, value: str = None):
        self.name = name
        self.value = value


class FilterKey(object):
    def __init__(self, rules: [] = None):
        self.rules = rules


class Filter(object):
    def __init__(self, key: FilterKey):
        self.key = key


class RocketMQConf(object):
    def __init__(self, instance_id: str = None, topic: str = None, access_key_id: str = None):
        self.instance_id = instance_id
        self.topic = topic
        self.access_key_id = access_key_id


class CloudFunctionConfiguration(object):
    def __init__(self, id: str = None, events: [] = None, filter: Filter = None, cloud_function: str = None):
        self.id = id
        self.events = events
        self.filter = filter
        self.cloud_function = cloud_function


class RocketMQConfiguration(object):
    def __init__(self, id: str = None, events: [] = None, filter: Filter = None, role: str = None,
                 rocket_mq: RocketMQConf = None):
        self.id = id
        self.events = events
        self.filter = filter
        self.role = role
        self.rocket_mq = rocket_mq


class PutBucketNotificationOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketNotificationOutput, self).__init__(resp)


class GetBucketNotificationOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketNotificationOutput, self).__init__(resp)
        data = resp.json_read()
        self.cloud_function_configurations = []
        self.rocket_mq_configurations = []
        cloud_functions = get_value(data, 'CloudFunctionConfigurations') or []
        rocket_mq_confs = get_value(data, 'RocketMQConfigurations') or []
        for function in cloud_functions:
            config = CloudFunctionConfiguration()
            config.id = get_value(function, 'RuleId')
            config.events = get_value(function, 'Events')
            config.cloud_function = get_value(function, 'CloudFunction')
            filter_json = get_value(function, 'Filter')
            if filter_json:
                config.filter = self._get_filter(filter_json)
            self.cloud_function_configurations.append(config)

        for rocket_mq_conf in rocket_mq_confs:
            config = RocketMQConfiguration()
            config.id = get_value(rocket_mq_conf, 'RuleId')
            config.events = get_value(rocket_mq_conf, 'Events')
            config.role = get_value(rocket_mq_conf, 'Role')
            rocket_mq_json = get_value(rocket_mq_conf, 'RocketMQ')
            if rocket_mq_json:
                rocket_mq = RocketMQConf(
                    instance_id=get_value(rocket_mq_json, 'InstanceId'),
                    topic=get_value(rocket_mq_json, 'Topic'),
                    access_key_id=get_value(rocket_mq_json, 'AccessKeyId'),
                )
                config.rocket_mq = rocket_mq
            filter_json = get_value(rocket_mq_conf, 'Filter')
            if filter_json:
                config.filter = self._get_filter(filter_json)
            self.rocket_mq_configurations.append(config)

    @staticmethod
    def _get_filter(filter_json):
        filter_key = FilterKey([])
        key_json = get_value(filter_json, 'TOSKey')
        if key_json:
            filter_rules = get_value(key_json, 'FilterRules') or []
            for rule in filter_rules:
                filter_key.rules.append(
                    FilterRule(name=get_value(rule, 'Name'), value=get_value(rule, 'Value')))
        return Filter(filter_key)


class CustomDomainRule(object):
    def __init__(self, cert_id: str = None, cert_status: CertStatus = None, domain: str = None, cname: str = None,
                 forbidden: bool = None, forbidden_reason: str = None):
        self.cert_id = cert_id
        self.cert_status = cert_status
        self.domain = domain
        self.cname = cname
        self.forbidden = forbidden
        self.forbidden_reason = forbidden_reason


class PutBucketCustomDomainOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketCustomDomainOutput, self).__init__(resp)


class ListBucketCustomDomainOutput(ResponseInfo):
    def __init__(self, resp):
        super(ListBucketCustomDomainOutput, self).__init__(resp)
        self.rules = []
        data = resp.json_read()
        custom_domain_rules = get_value(data, 'CustomDomainRules') or []
        for custom_domain_rule in custom_domain_rules:
            self.rules.append(
                CustomDomainRule(cert_id=get_value(custom_domain_rule, 'CertId'),
                                 domain=get_value(custom_domain_rule, 'Domain'),
                                 cname=get_value(custom_domain_rule, 'Cname'),
                                 forbidden=get_value(custom_domain_rule, 'Forbidden'),
                                 forbidden_reason=get_value(custom_domain_rule, 'ForbiddenReason'),
                                 cert_status=get_value(custom_domain_rule, 'CertStatus',
                                                       lambda x: convert_cert_status(x))))


class DeleteCustomDomainOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteCustomDomainOutput, self).__init__(resp)


class AccessLogConfiguration(object):
    def __init__(self, use_service_topic: bool = None, tls_project_id: str = None, tls_topic_id: str = None):
        self.use_service_topic = use_service_topic
        self.tls_project_id = tls_project_id
        self.tls_topic_id = tls_topic_id


class RealTimeLogConfiguration(object):
    def __init__(self, role: str = None, configuration: AccessLogConfiguration = None):
        self.role = role
        self.configuration = configuration


class PutBucketRealTimeLogOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketRealTimeLogOutput, self).__init__(resp)


class GetBucketRealTimeLog(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketRealTimeLog, self).__init__(resp)
        self.configuration = RealTimeLogConfiguration()
        data = resp.json_read()
        bucket_real_time_log = get_value(data, 'RealTimeLogConfiguration')
        self.configuration.role = get_value(bucket_real_time_log, 'Role')
        access_log = get_value(bucket_real_time_log, 'AccessLogConfiguration')
        if access_log:
            self.configuration.configuration = AccessLogConfiguration()
            self.configuration.configuration.use_service_topic = get_value(access_log, 'UseServiceTopic')
            self.configuration.configuration.tls_topic_id = get_value(access_log, 'TLSTopicID')
            self.configuration.configuration.tls_project_id = get_value(access_log, 'TLSProjectID')


class DeleteBucketRealTimeLog(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketRealTimeLog, self).__init__(resp)


class ResumableCopyObjectOutput(object):
    def __init__(self, resp: CompleteMultipartUploadOutput, ssec_algorithm, ssec_key_md5, encoding_type):
        self.request_id = resp.request_id
        self.id2 = resp.id2
        self.status_code = resp.status_code
        self.header = resp.header
        self.bucket = resp.bucket
        self.key = resp.key
        self.upload_id = resp.request_id
        self.etag = resp.etag
        self.location = resp.location
        self.version_id = resp.version_id
        self.hash_crc64_ecma = resp.hash_crc64_ecma
        self.ssec_algorithm = ssec_algorithm
        self.ssec_key_md5 = ssec_key_md5
        self.encoding_type = encoding_type


class PreSignedPolicyURlInputOutput(object):
    def __init__(self, signed_query, host, scheme, bucket=None):
        self.signed_query = signed_query
        self._host = host
        self._scheme = scheme
        self._bucket = bucket

    def get_signed_url_for_list(self, additional_query=None) -> str:
        if additional_query is None:
            return _make_virtual_host_url(self._host, self._scheme, self._bucket, '') + '?' + self.signed_query

        return _make_virtual_host_url(self._host, self._scheme, self._bucket,
                                      '') + '?' + self.signed_query + '&' + '&'.join(
            _param_to_quoted_query(k, v) for k, v in additional_query.items())

    def get_signed_url_for_get_or_head(self, key: str, additional_query=None) -> str:
        if additional_query is None:
            return _make_virtual_host_url(self._host, self._scheme, self._bucket, key) + '?' + self.signed_query

        return _make_virtual_host_url(self._host, self._scheme, self._bucket,
                                      key) + '?' + self.signed_query + '&' + '&'.join(
            _param_to_quoted_query(k, v) for k, v in additional_query.items())


class CopyPartInfo(object):
    def __init__(self, part_number, copy_source_range_start, copy_source_range_end, etag=None):
        self.part_number = part_number
        self.copy_source_range_start = copy_source_range_start
        self.copy_source_range_end = copy_source_range_end
        self.etag = etag

    def __str__(self):
        info = {'part_number': self.part_number,
                'copy_source_range_start': self.copy_source_range_start,
                'copy_source_range_end': self.copy_source_range_end,
                'etag': self.etag}
        return str(info)


class RestoreJobParameters(object):
    def __init__(self, tier: TierType):
        self.tier = tier


class RestoreObjectOutput(ResponseInfo):
    def __init__(self, resp):
        super(RestoreObjectOutput, self).__init__(resp)


class RenameObjectOutput(ResponseInfo):
    def __init__(self, resp):
        super(RenameObjectOutput, self).__init__(resp)


class GetBucketRenameOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketRenameOutput, self).__init__(resp)
        data = resp.json_read()
        self.rename_enable = get_value(data, 'RenameEnable', bool)


class PutBucketRenameOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketRenameOutput, self).__init__(resp)


class DeleteBucketRenameOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketRenameOutput, self).__init__(resp)


class PutBucketTaggingOutput(ResponseInfo):
    def __init__(self, resp):
        super(PutBucketTaggingOutput, self).__init__(resp)


class GetBucketTaggingOutput(ResponseInfo):
    def __init__(self, resp):
        super(GetBucketTaggingOutput, self).__init__(resp)
        self.tag_set = []
        data = resp.json_read()
        tag_set = get_value(data, 'TagSet')
        if tag_set:
            tags = get_value(tag_set, 'Tags')
            if tags:
                for tag in tags:
                    self.tag_set.append(Tag(
                        get_value(tag, 'Key'),
                        get_value(tag, 'Value')
                    ))


class DeleteBucketTaggingOutput(ResponseInfo):
    def __init__(self, resp):
        super(DeleteBucketTaggingOutput, self).__init__(resp)



import logging

from . import exceptions, models2, log
from .auth import Auth, FederationAuth
from .client import TosClient
from .clientv2 import TosClientV2
from .enum import *
from .utils import RateLimiter, to_bytes, DnsCacheService
from .credential import StaticCredentialsProvider, EcsCredentialsProvider, EnvCredentialsProvider

logger = logging.getLogger('tos')


def set_logger(file_path='TosClient.log', name="tos", level=log.INFO, format_string=None):
    global logger
    if not format_string:
        format_string = "%(asctime)s %(name)s [%(levelname)s] %(thread)d : %(message)s"
    logger = logging.getLogger(name)
    logger.setLevel(level)
    fh = logging.FileHandler(file_path)
    fh.setLevel(level)
    formatter = logging.Formatter(format_string)
    fh.setFormatter(formatter)
    logger.addHandler(fh)




from requests.structures import CaseInsensitiveDict

from .utils import get_etag, get_value, parse_gmt_time_to_utc_datetime


class RequestResult(object):
    def __init__(self, resp):
        self.resp = resp
        self.status = resp.status
        self.headers = resp.headers
        self.request_id = resp.request_id
        self.version_id = get_value(self.headers, 'x-tos-version-id')
        self.delete_marker = get_value(self.headers, 'x-tos-delete-marker', bool)


class CreateBucketResult(RequestResult):
    def __init__(self, resp):
        super(CreateBucketResult, self).__init__(resp)
        self.location = get_value(self.headers, "Location")


class HeadBucketResult(RequestResult):
    def __init__(self, resp):
        super(HeadBucketResult, self).__init__(resp)
        self.region = get_value(self.headers, "x-tos-bucket-region")


class CreateMultipartUploadResult(RequestResult):
    def __init__(self, resp):
        super(CreateMultipartUploadResult, self).__init__(resp)
        self.upload_id = None
        self.bucket = None
        self.key = None

        self.sse_customer_algorithm = get_value(self.headers, "x-tos-server-side-encryption-customer-algorithm")
        self.sse_customer_key = get_value(self.headers, "x-tos-server-side-encryption-customer-key")
        self.sse_customer_key_md5 = get_value(self.headers, "x-tos-server-side-encryption-customer-key-md5")


class HeadObjectResult(RequestResult):
    def __init__(self, resp):
        super(HeadObjectResult, self).__init__(resp)
        self.content_type = get_value(self.headers, "content-type")
        self.content_length = get_value(self.headers, "content-length", int)
        self.etag = get_etag(self.headers)

        self.sse_customer_algorithm = get_value(self.headers, "x-tos-server-side-encryption-customer-algorithm")
        self.sse_customer_key = get_value(self.headers, "x-tos-server-side-encryption-customer-key")
        self.sse_customer_key_md5 = get_value(self.headers, "x-tos-server-side-encryption-customer-key-md5")

        self.metadata = CaseInsensitiveDict()
        for k in self.headers:
            if k.startswith('x-tos-meta-'):
                self.metadata[k[11:]] = self.headers[k]

        self.last_modified = get_value(self.headers, 'last-modified')
        if self.last_modified:
            self.last_modified = parse_gmt_time_to_utc_datetime(self.last_modified)
        self.expires = get_value(self.headers, 'expires')
        if self.expires:
            self.expires = parse_gmt_time_to_utc_datetime(self.expires)


class CopyObjectResult(RequestResult):
    def __init__(self, resp):
        super(CopyObjectResult, self).__init__(resp)
        self.etag = ''
        self.last_modified = ''
        self.copy_source_version_id = get_value(self.headers, "x-tos-copy-source-version-id")


class UploadPartCopyResult(RequestResult):
    def __init__(self, resp):
        super(UploadPartCopyResult, self).__init__(resp)
        self.etag = ''
        self.last_modified = ''
        self.copy_source_version_id = get_value(self.headers, "x-tos-copy-source-version-id")

        self.sse_customer_algorithm = get_value(self.headers, "x-tos-server-side-encryption-customer-algorithm")
        self.sse_customer_key = get_value(self.headers, "x-tos-server-side-encryption-customer-key")
        self.sse_customer_key_md5 = get_value(self.headers, "x-tos-server-side-encryption-customer-key-md5")


class GetObjectResult(HeadObjectResult):
    def __init__(self, resp):
        super(GetObjectResult, self).__init__(resp)
        self.content_range = get_value(self.headers, "content-range")
        self.stream = self.resp

    def read(self, amt=None):
        return self.stream.read(amt)

    def __iter__(self):
        return iter(self.stream)


class PutObjectResult(RequestResult):
    def __init__(self, resp):
        super(PutObjectResult, self).__init__(resp)
        self.etag = get_etag(self.headers)

        self.sse_customer_algorithm = get_value(self.headers, "x-tos-server-side-encryption-customer-algorithm")
        self.sse_customer_key = get_value(self.headers, "x-tos-server-side-encryption-customer-key")
        self.sse_customer_key_md5 = get_value(self.headers, "x-tos-server-side-encryption-customer-key-md5")


class AppendObjectResult(RequestResult):
    def __init__(self, resp):
        super(AppendObjectResult, self).__init__(resp)
        self.etag = get_etag(self.headers)
        self.sse_customer_algorithm = get_value(self.headers, "x-tos-server-side-encryption-customer-algorithm")
        self.sse_customer_key = get_value(self.headers, "x-tos-server-side-encryption-customer-key")
        self.sse_customer_key_md5 = get_value(self.headers, "x-tos-server-side-encryption-customer-key-md5")


class ListBucketResult(RequestResult):
    def __init__(self, resp):
        super(ListBucketResult, self).__init__(resp)
        self.bucket_list = []  # BucketInfo list
        self.Owner = None  # UserInfo


class ListMultipartUploadsResult(RequestResult):
    def __init__(self, resp):
        super(ListMultipartUploadsResult, self).__init__(resp)
        self.bucket = ''
        self.upload_id_marker = ''
        self.next_key_marker = ''
        self.next_upload_id_marker = ''
        self.delimiter = ''
        self.prefix = ''
        self.max_uploads = None
        self.is_truncated = False
        self.upload_list = []  # MultipartUploadInfo list
        self.common_prefix_list = []  # CommonPrefixInfo list


class ListPartsResult(RequestResult):
    def __init__(self, resp):
        super(ListPartsResult, self).__init__(resp)
        self.bucket = ''
        self.key = ''
        self.upload_id = ''
        self.part_number_marker = None
        self.next_part_number_marker = None
        self.max_parts = None
        self.is_truncated = False
        self.storage_class = ''
        self.owner = None  # UserInfo list
        self.part_list = []  # UploadPartInfo list


class CompleteMultipartUploadResult(RequestResult):
    def __init__(self, resp):
        super(CompleteMultipartUploadResult, self).__init__(resp)
        self.bucket = ''
        self.location = ''
        self.key = ''
        self.etag = ''


class ListObjectsResult(RequestResult):
    def __init__(self, resp):
        super(ListObjectsResult, self).__init__(resp)
        self.name = ''
        self.prefix = ''
        self.key_count = None
        self.marker = ''
        self.max_keys = None
        self.next_marker = ''
        self.delimiter = ''
        self.is_truncated = False
        self.encoding_type = 'url'
        self.common_prefix_list = []  # CommonPrefixInfo list
        self.object_list = []  # ObjectInfo list



class ListObjectVersionsResult(RequestResult):
    def __init__(self, resp):
        super(ListObjectVersionsResult, self).__init__(resp)
        self.name = ''
        self.prefix = ''
        self.key_marker = ''
        self.version_id_marker = ''
        self.max_keys = None
        self.next_key_marker = ''
        self.next_version_id_marker = ''
        self.delimiter = ''
        self.is_truncated = False
        self.encoding_type = 'url'
        self.common_prefix_list = []  # string list
        self.version_list = []  # ObjectVersionInfo list
        self.delete_marker_list = []  # DeleteMarker list


class DeleteObjectsResult(RequestResult):
    def __init__(self, resp):
        super(DeleteObjectsResult, self).__init__(resp)
        self.deleted_list = []  # DeletedObjectInfo list
        self.error_list = []  # DeletedErrInfo list


class GetObjectAclResult(RequestResult):
    def __init__(self, resp):
        super(GetObjectAclResult, self).__init__(resp)
        self.grant_list = []  # GrantInfo list
        self.owner = None


# result 中的元素信息
class UserInfo(object):
    def __init__(self, id, name):
        self.id = id
        self.name = name


class BucketInfo(object):
    def __init__(self, name, location, creation_date, extranet_endpoint, intranet_endpoint):
        self.name = name
        self.location = location
        self.creation_date = creation_date
        self.extranet_endpoint = extranet_endpoint
        self.intranet_endpoint = intranet_endpoint


class GranteeInfo(object):
    def __init__(self, id, display_name, type, canned):
        self.id = id
        self.display_name = display_name
        self.type = type
        self.canned = canned


class GrantInfo(object):
    def __init__(self, id, display_name, type, canned, permission):
        self.grantee = GranteeInfo(id, display_name, type, canned)
        self.permission = permission


class MultipartUploadInfo(object):
    def __init__(self, key, upload_id, storage_class, initiated, owner=None):
        self.key = key
        self.upload_id = upload_id
        self.storage_class = storage_class
        self.initiated = initiated
        self.owner = owner


class UploadPartInfo(object):
    def __init__(self, part_number, last_modified, etag, size):
        self.part_number = part_number
        self.last_modified = last_modified
        self.etag = etag
        self.size = size


class CommonPrefixInfo(object):
    def __init__(self, prefix):
        self.prefix = prefix


class ObjectInfo(object):
    def __init__(self, key, last_modified, etag, size, storage_class, owner=None):
        self.key = key
        self.last_modified = last_modified
        self.etag = etag
        self.size = size
        self.storage_class = storage_class
        self.owner = owner


class ObjectVersionInfo(object):
    def __init__(self, key, is_latest, last_modified, etag, size, storage_class, version_id, owner=None):
        self.key = key
        self.is_latest = is_latest
        self.last_modified = last_modified
        self.etag = etag
        self.size = size
        self.storage_class = storage_class
        self.version_id = version_id
        self.owner = owner


class DeleteMarkerInfo(object):
    def __init__(self, key, is_latest, last_modified, version_id, owner=None):
        self.key = key
        self.is_latest = is_latest
        self.last_modified = last_modified
        self.version_id = version_id
        self.owner = owner


class DeletedObjectInfo(object):
    def __init__(self, key, version_id, delete_marker, delete_marker_version_id):
        self.key = key
        self.version_id = version_id
        self.delete_marker = delete_marker
        self.delete_marker_version_id = delete_marker_version_id


class DeletedErrInfo(object):
    def __init__(self, code, message, key, version_id):
        self.code = code
        self.message = message
        self.key = key
        self.version_id = version_id
